import os
import errno
import threading
import datetime
import tempfile
import shutil
import json
import time
import socket
from copy import deepcopy
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.GPIDev.Credentials import credential_store
import GangaCore.Utility.execute as gexecute
logger = getLogger()

# Cache
# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
DIRAC_ENV = {}
DIRAC_INCLUDE = ''
Dirac_Env_Lock = threading.Lock()
Dirac_Proxy_Lock = threading.Lock()
Dirac_Exec_Lock = threading.Lock()
# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class GangaDiracError(GangaException):
    """ Exception type which is thrown from problems executing a command against DIRAC """
    def __init__(self, message, dirac_id = None, job_id = None):
        GangaException.__init__(self, message)
        self.dirac_id = dirac_id
        self.job_id = job_id
    def __str__(self):
        if self.job_id and self.dirac_id:
            return "GangaDiracError, Job %s with Dirac ID %s : %s" % (self.job_id, self.dirac_id, self.message)
        else:
            return GangaException.__str__(self)


def getDiracEnv(sourceFile = None):
    """
    Returns the dirac environment stored in a global dictionary by GangaCore.
    Once loaded and stored this is used for executing all DIRAC code in future
    Args:
        sourceFile (str): This is an optional file path which points to the env which should be sourced for this DIRAC
    """
    global DIRAC_ENV
    with Dirac_Env_Lock:
        if sourceFile is None:
            sourceFile = 'default'
            cache_file = getConfig('DIRAC')['DiracEnvJSON']
            source_command = getConfig('DIRAC')['DiracEnvSource']
            if not cache_file and not source_command:
                source_command = getConfig('DIRAC')['DiracEnvFile']
        else:
            # Needed for backwards compatibility with old configs...
            cache_file = None
            source_command = sourceFile

        if sourceFile not in DIRAC_ENV:
            if cache_file:
                DIRAC_ENV[sourceFile] = read_env_cache(cache_file)
            elif source_command:
                DIRAC_ENV[sourceFile] = get_env(source_command)
            else:
                logger.error("'DiracEnvSource' config variable empty")
                logger.error("%s  %s" % (getConfig('DIRAC')['DiracEnvJSON'], getConfig('DIRAC')['DiracEnvSource']))

    return DIRAC_ENV[sourceFile]


def get_env(env_source):
    """
    Given a source command, return the DIRAC environment that the
    command created.

    Args:
        env_source: a command which can be sourced, providing the desired environment

    Returns:
        dict: the environment

    """
    logger.debug('Running DIRAC source command %s', env_source)
    env = dict(os.environ)
    gexecute.execute('source {0}'.format(env_source), shell=True, env=env, update_env=True)
    if not any(key.startswith('DIRAC') for key in env):
        fake_dict = {}
        with open(env_source) as _env:
            for _line in _env.readlines():
                split_val = _line.split('=')
                if len(split_val) == 2:
                    key = split_val[0]
                    value = split_val[1]
                    fake_dict[key] = value
        if not any(key.startswith('DIRAC') for key in fake_dict):
            logger.error("Env: %s" % str(env))
            logger.error("Fake: %s" % str(fake_dict))
            raise RuntimeError("'DIRAC*' not found in environment")
        else:
            return fake_dict
    return env


def write_env_cache(env, cache_filename):
    """
    Given a command and a file path, source the command and store it
    in the file

    Args:
        env (dict): the environment
        cache_filename: a full path to a file to store the cache in

    """
    cache_dir = os.path.dirname(cache_filename)
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    with open(cache_filename, 'w') as cache_file:
        json.dump(env, cache_file)


def read_env_cache(cache_filename):
    """
    Args:
        cache_filename: a full path to a file to store the cache in

    Returns:
        dict: the cached environment

    """
    logger.debug('Reading DIRAC cache file at %s', cache_filename)
    with open(cache_filename, 'r') as cache_file:
        env = json.load(cache_file)
    # Convert unicode strings to byte strings
    env = dict((k.encode('utf-8'), v.encode('utf-8')) for k, v in env.items())
    return env

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def getDiracCommandIncludes(force=False):
    """
    This helper function returns the Ganga DIRAC helper functions which are called by Ganga code to talk to DIRAC
    These are loaded from disk once and then saved in memory.
    Args:
        force (bool): Triggers a reload from disk when True
    """
    global DIRAC_INCLUDE
    with Dirac_Env_Lock:
        if DIRAC_INCLUDE == '' or force:
            for fname in getConfig('DIRAC')['DiracCommandFiles']:
                if not os.path.exists(fname):
                    raise RuntimeError("Specified Dirac command file '%s' does not exist." % fname)
                with open(fname, 'r') as inc_file:
                    DIRAC_INCLUDE += inc_file.read() + '\n'

    return DIRAC_INCLUDE

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def getValidDiracFiles(job, names=None):
    """
    This is a generator for all DiracFiles in a jobs outputfiles
    TODO: Is this still called anywhere?
    Args:
        job (Job): The job which is having it's DiracFiles tested
        names (list): list of strings of names to be matched to namePatterns in outputfiles
    """
    from GangaDirac.Lib.Files.DiracFile import DiracFile
    if job.subjobs:
        for sj in job.subjobs:
            for df in (f for f in sj.outputfiles if isType(f, DiracFile)):
                if df.subfiles:
                    for valid_sf in (sf for sf in df.subfiles if sf.lfn != '' and (names is None or sf.namePattern in names)):
                        yield valid_sf
                else:
                    if df.lfn != '' and (names is None or df.namePattern in names):
                        yield df
    else:
        for df in (f for f in job.outputfiles if isType(f, DiracFile)):
            if df.subfiles:
                for valid_sf in (sf for sf in df.subfiles if sf.lfn != '' and (names is None or sf.namePattern in names)):
                    yield valid_sf
            else:
                if df.lfn != '' and (names is None or df.namePattern in names):
                    yield df


def execute(command,
            timeout=getConfig('DIRAC')['Timeout'],
            env=None,
            cwd=None,
            shell=False,
            python_setup='',
            eval_includes=None,
            update_env=False,
            return_raw_dict=False,
            cred_req=None,
            new_subprocess = False
            ):
    """
    Execute a command on the local DIRAC server.

    This function blocks until the server returns.
    
    Args:
        command (str): This is the command we're running within our DIRAC session
        timeout (int): This is the length of time that a DIRAC call has before it's decided some interaction has timed out
        env (dict): an optional environment to execute the DIRAC code in
        cwd (str): an optional string to a valid path where this code should be executed
        shell (bool): Should this code be executed in a new shell environment
        python_setup (str): Optional extra code to pass to python when executing
        eval_includes (???): TODO document me
        update_env (bool): Should this modify the given env object with the env after the command has executed
        return_raw_dict(bool): Should we return the raw dict from the DIRAC interface or parse it here
        cred_req (ICredentialRequirement): What credentials does this call need
        new_subprocess(bool): Do we want to do this in a fresh subprocess or just connect to the DIRAC server process?
    """

    if cwd is None:
        # We can in all likelyhood be in a temp folder on a shared (SLOW) filesystem
        # If we are we do NOT want to execute commands which will involve any I/O on the system that isn't needed
        cwd_ = tempfile.mkdtemp()
    else:
        # We know were whe want to run, lets just run there
        cwd_ = cwd

    from GangaDirac.BOOT import startDiracProcess
    returnable = ''
    if not new_subprocess:
        with Dirac_Exec_Lock:
            # First check if a Dirac process is running
            from GangaDirac.BOOT import running_dirac_process
            if not running_dirac_process:
                startDiracProcess()
            #Set up a socket to connect to the process
            from GangaDirac.BOOT import dirac_process_ids
            HOST = 'localhost'  # The server's hostname or IP address
            PORT = dirac_process_ids[1]        # The port used by the server

            #Put inside a try/except in case the existing process has timed out
            try:
                s= socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.connect((HOST, PORT))
            except socket.error as serr:
                #Start a new process
                startDiracProcess()
                from GangaDirac.BOOT import dirac_process_ids
                PORT = dirac_process_ids[1]
                s= socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.connect((HOST, PORT))

            #Send a random string, then change the directory to carry out the command, then send the command
            command_to_send  = str(dirac_process_ids[2])
            command_to_send += 'os.chdir("%s")\n' % cwd_
            command_to_send += command
            s.sendall(b'%s###END-TRANS###' % command_to_send)
            out = ''
            while '###END-TRANS###' not in out:
                data = s.recv(1024)
                out += data
            s.close()
            returnable = eval(out)

    else:
        if env is None:
            if cred_req is None:
                env = getDiracEnv()
            else:
                env = getDiracEnv(cred_req.dirac_env)
        if python_setup == '':
            python_setup = getDiracCommandIncludes()

        if cred_req is not None:
            env['X509_USER_PROXY'] = credential_store[cred_req].location
            if os.getenv('KRB5CCNAME'):
                env['KRB5CCNAME'] = os.getenv('KRB5CCNAME')

        returnable = gexecute.execute(command,
                                      timeout=timeout,
                                      env=env,
                                      cwd=cwd_,
                                      shell=shell,
                                      python_setup=python_setup,
                                      eval_includes=eval_includes,
                                      update_env=update_env)

        # If the time 
        if returnable == 'Command timed out!':
            raise GangaDiracError("DIRAC command timed out")

        # TODO we would like some way of working out if the code has been executed correctly
        # Most commands will be OK now that we've added the check for the valid proxy before executing commands here

    if cwd is None:
        shutil.rmtree(cwd_, ignore_errors=True)

    if isinstance(returnable, dict) and not return_raw_dict:
        # If the output is a dictionary allow for automatic error detection
        if returnable['OK']:
            return returnable['Value']
        else:
            raise GangaDiracError(returnable['Message'])
    elif isinstance(returnable, dict):
        # If the output is a dictionary return and it has been requested, then return it
        return returnable
    else:
        # Else raise an exception as it should be a dictionary
        raise  GangaDiracError(returnable)




##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# Copyright (C) 2003 The Ganga Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
##########################################################################

import os
import os.path
import string
import sys
import time
import re

from GangaCore import _gangaVersion, _gangaPythonPath, _development
from GangaCore.Utility.Config.Config import getConfig
from GangaCore.Utility import stacktracer
import GangaCore.Runtime

from IPython.terminal.prompts import Prompts, Token
class GangaPrompt(Prompts):
    environment = None
        # Here __flushCmd evaluates the object in the local_ns with this name to a str
        # The magic function time here is the builtin time prompt This renders in the same way as:
        #
        # -------------------------------------
        #
        # [18:45:20]
        # Ganga In [1]: if True:
        #          ...:     print("Hello")
        #          ...:
        #               Hello
        #
        # [18:45:43]
        # Ganga In [2]:
        #
        # -------------------------------------

    def in_prompt_tokens(self, cli = None):
        from datetime import datetime
        from GangaCore.Utility.logging import flushAtIPythonPrompt
        __flushCmd = flushAtIPythonPrompt()
        return [(Token.Prompt,  str(__flushCmd)+'['+str(datetime.now().strftime("%H:%M:%S"))+']\nGanga In ['.format(GangaPrompt.environment)),
                (Token.PromptNum, str(self.shell.execution_count)),
                (Token.Prompt, ']: ')]

    def continuation_prompt_tokens(self, cli = None, width = 4):
        return [(Token.Prompt, '         ...: ')]

    def out_prompt_tokens(self, cli = None):
        return [(Token.OutPrompt, 'Ganga Out ['+str(self.shell.execution_count)+']: ')]

def new_version_format_to_old(version):
    """
    Convert from 'x.y.z'-style format to 'Ganga-x-y-z'

    Example:
        >>> new_version_format_to_old('6.1.11')
        'Ganga-6-1-11'
    """
    return 'Ganga-'+version.replace('.', '-')



from GangaCore.Utility.files import fullpath

# This code can help debugging when files aren't closed correctly and
# managing I/O

DEBUGFILES = False
MONITOR_FILES = False

if DEBUGFILES or MONITOR_FILES:
    import __builtin__
    openfiles = {}
    oldfile = __builtin__.file

    class newfile(oldfile):

        def __init__(self, *args):
            self.x = args[0]
            if DEBUGFILES:
                from GangaCore.Utility.logging import getLogger
                logger = getLogger(modulename=True)
                logger.debug("init")
                logger.debug("### OPENING %s ###" % self.x)
            oldfile.__init__(self, *args)
            openfiles[self.x] = self

        def close(self):
            if DEBUGFILES:
                from GangaCore.Utility.logging import getLogger
                logger = getLogger(modulename=True)
                logger.debug("### CLOSING %s ###" % self.x)
            oldfile.close(self)
            #openfiles[ self.x ] = None
            del openfiles[self.x]

    oldopen = __builtin__.open

    def newopen(*args):
        if DEBUGFILES:
            from GangaCore.Utility.logging import getLogger
            logger = getLogger()
            logger.debug("NewOpen")
        return newfile(*args)
    __builtin__.file = newfile
    __builtin__.open = newopen

    def printOpenFiles():
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("### %d OPEN FILES: [%s]" % (len(openfiles), ", ".join(
            str(f) for f in openfiles.keys() if openfiles[f] is not None)))

    safeFiles = ['.ganga.log', '.gangarc', 'ipythonrc', 'history', 'persist']

    def safeCloseOpenFiles():
        for f in openfiles.keys():
            if f not in safeFiles:
                openfiles[f].close()


def manualExportToGPI(my_interface=None):
    """ Additionally populate the GPI elements which are required
        This is here rather than in utility due to having knowledge of what is inside Ganga
        whilst Utility should not have too much (or any) of this
    """

    if not my_interface:
        import GangaCore.GPI
        my_interface = GangaCore.GPI

    from GangaCore.Runtime.GPIexport import exportToInterface

    from GangaCore.GPIDev.Base import ProtectedAttributeError, ReadOnlyObjectError, GangaAttributeError
    from GangaCore.GPIDev.Lib.Job.Job import JobError

    exportToInterface(my_interface, 'GangaAttributeError', GangaAttributeError, 'Exceptions')
    exportToInterface(my_interface, 'ProtectedAttributeError', ProtectedAttributeError, 'Exceptions')
    exportToInterface(my_interface, 'ReadOnlyObjectError', ReadOnlyObjectError, 'Exceptions')
    exportToInterface(my_interface, 'JobError', JobError, 'Exceptions')

    from GangaCore.Runtime.GPIFunctions import license, typename, categoryname, plugins, convert_merger_to_postprocessor

    exportToInterface(my_interface, 'license', license, 'Functions')
    # FIXME:
    #from GangaCore.Runtime.GPIFunctions import applications, backends, list_plugins
    #exportToInterface(my_interface, 'applications', applications, 'Functions)
    #exportToInterface(my_interface, 'backends', backends, 'Functions')
    #exportToInterface(my_interface, 'list_plugins', list_plugins, 'Functions')
    # FIXME: END DEPRECATED

    from GangaCore.GPIDev.Credentials import credential_store
    exportToInterface(my_interface, 'credential_store', credential_store, 'Objects', 'Credential store')

    exportToInterface(my_interface, 'typename', typename, 'Functions')
    exportToInterface(my_interface, 'categoryname', categoryname, 'Functions')
    exportToInterface(my_interface, 'plugins', plugins, 'Functions')
    exportToInterface(my_interface, 'convert_merger_to_postprocessor', convert_merger_to_postprocessor, 'Functions')

    from GangaCore.GPIDev.Persistency import export, load
    exportToInterface(my_interface, 'load', load, 'Functions')
    exportToInterface(my_interface, 'export', export, 'Functions')

    # export full_print
    from GangaCore.GPIDev.Base.VPrinter import full_print
    exportToInterface(my_interface, 'full_print', full_print, 'Functions')

    import GangaCore.GPIDev.Lib.Config
    exportToInterface(my_interface, 'config', GangaCore.GPIDev.Lib.Config.config, 'Objects', 'access to Ganga configuration')
    exportToInterface(my_interface, 'ConfigError', GangaCore.GPIDev.Lib.Config.ConfigError, 'Exceptions')

    from GangaCore.Utility.feedback_report import report
    exportToInterface(my_interface, 'report', report, 'Functions')

    from GangaCore.Core.InternalServices.Coordinator import enableInternalServices, disableMonitoringService, enableMonitoringService, disableInternalServices
    exportToInterface(my_interface, 'reactivate', enableInternalServices, 'Functions')
    exportToInterface(my_interface, 'disableMonitoring', disableMonitoringService, 'Functions')
    exportToInterface(my_interface, 'enableMonitoring', enableMonitoringService, 'Functions')
    exportToInterface(my_interface, 'disableServices', disableInternalServices, 'Functions')

    from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
    exportToInterface(my_interface, 'GangaList', GangaList, 'Classes')

    from GangaCore.GPIDev.Lib.Registry.JobRegistry import jobSlice
    exportToInterface(my_interface, "jobSlice", jobSlice, "Functions")

class GangaProgram(object):

    """ High level API to create instances of Ganga programs and configure/run it """

    def __init__(self, hello_string=None, argv=sys.argv):
        """ make an instance of Ganga program
        use default hello_string if not specified
        use sys.argv as arguments if not specified"""

        self.argv = argv[:]
        # record the start time.Currently we are using this in performance measurements
        # see Ganga/old_test/Performance tests
        self.start_time = time.time()

        versionString = _gangaVersion
        if _development:
            versionString += ' - DEV'

        if hello_string is None:
            self.hello_string = """
*** Welcome to Ganga ***
Version: %s
Documentation and support: http://cern.ch/ganga
Type help() or help('index') for online help.

This is free software (GPL), and you are welcome to redistribute it
under certain conditions; type license() for details.

""" % versionString
        else:
            self.hello_string = hello_string

        # by default enter interactive mode
        self.interactive = True

        self.default_config_file = os.path.expanduser('~/.gangarc')

        if sys.hexversion < 0x020700F0:
            version = '{0}.{1}'.format(sys.version_info[0], sys.version_info[1])
            from GangaCore.Utility.logging import getLogger
            logger = getLogger()
            logger.warning('Ganga will soon be depending on Python 2.7. '
                           'You have Python {version} installed. '
                           'See https://github.com/ganga-devs/ganga/wiki/Python-2.7'.format(version=version))

    def exit(self, *msg):
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()
        logger.info(self.hello_string)
        for m in msg:
            logger.error(m)
        sys.exit(1)

    # parse the options

    def parseOptions(self):
        from optparse import OptionParser

        usage = self.hello_string + \
            """\nusage: %prog [options] [script] [args] ..."""

        parser = OptionParser(usage, version=_gangaVersion)

        parser.add_option("-i", dest="force_interactive", action="store_true",
                          help='enter interactive mode after running script')

        parser.add_option("--webgui", dest="webgui",  action="store_true", default='False',
                          help='starts web GUI monitoring server')

        parser.add_option("--config", dest="config_file", action="store", metavar="FILE", default=None,
                          help='read user configuration from FILE, overrides the GANGA_CONFIG_FILE environment variable. Default: ~/.gangarc')

        parser.add_option("--config-path", dest='config_path', action="store", default=None,
                          help='site/experiment config customization path, overrides the GANGA_CONFIG_PATH environment variable. The relative paths are resolved wrt to the release directory. To use a specific file you should specify the absolute path. Default: None')

        parser.add_option("-g", "--generate-config", dest='generate_config', action="store_const", const=1,
                          help='generate a default config file, backup the existing one')

        parser.add_option("-o", "--option", dest='cmdline_options', action="append", default=[], metavar='EXPR',
                          help='set configuration options, may be repeated mutiple times,'
                               'for example: -o[Logging]GangaCore.Lib=DEBUG -oGangaLHCb=INFO -o[Configuration]TextShell = IPython '
                               'The VALUE of *_PATH option is prepended. To reset it use :::VALUE')

        parser.add_option("--quiet", dest="force_loglevel", action="store_const", const='ERROR',
                          help='only ERROR messages are printed')

        parser.add_option("--very-quiet", dest="force_loglevel", action="store_const", const='CRITICAL',
                          help='only CRITICAL messages are printed')

        parser.add_option("--debug", dest="force_loglevel", action="store_const", const='DEBUG',
                          help='all messages including DEBUG are printed')

        parser.add_option("--no-mon", dest='monitoring', action="store_const", const=0,
                          help='disable the monitoring loop (useful if you run multiple Ganga sessions)')

        parser.add_option("--no-prompt", dest='prompt', action="store_const", const=0,
                          help='never prompt interactively for anything except IPython (FIXME: this is not fully implemented)')

        parser.add_option("--no-rexec", dest="rexec", action="store_const", const=0,
                          help='[DEPRECATED] rely on existing environment and do not re-exec ganga process'
                               'to setup runtime plugin modules (affects LD_LIBRARY_PATH)')

        parser.add_option("--test", dest='TEST', action="store_true", default=False,
                          help='run Ganga test(s) using internal test-runner. It requires GangaTest package to be installed.'
                               'Usage example: *ganga --test Ganga/old_test/MyTestcase* .'
                               'Refer to [TestingFramework] section in Ganga config for more information on how to configure the test runner.')

        parser.add_option("--daemon", dest='daemon', action="store_true", default=False,
                          help='run Ganga as service.')

        parser.set_defaults(force_interactive=False, config_file=None,
                            force_loglevel=None, rexec=1, monitoring=1, prompt=1, generate_config=None)
        parser.disable_interspersed_args()

        (self.options, self.args) = parser.parse_args(args=self.argv[1:])

        # check for --no-rexec. It does nothing now!
        if self.options.rexec == 0:
            from GangaCore.Utility.logging import getLogger
            getLogger().warning("Ganga no longer re-execs. --no-rexec option will be ignored.")

        def file_opens(f, message):
            try:
                return open(f)
            except IOError as x:
                self.exit(message, x)

        if self.options.config_file == '':
            self.options.config_file = None
        self.options.config_file_set_explicitly = not self.options.config_file is None

        # use GANGA_CONFIG_FILE env var if it's set
        if self.options.config_file is None:
            self.options.config_file = os.environ.get(
                'GANGA_CONFIG_FILE', None)

        if self.options.config_file:
            import GangaCore.Utility.files
            self.options.config_file = GangaCore.Utility.files.expandfilename(
                self.options.config_file)
            open_file = file_opens(
                self.options.config_file, 'reading configuration file')
            open_file.close()
        # we run in the batch mode if a script has been specified and other
        # options (such as -i) do not force it
        if len(self.args) > 0:
            if not self.options.force_interactive:
                self.interactive = False

# Can't check here if the file is readable, because the path isn't known
#           file_opens(self.args[0],'reading script')

    @staticmethod
    def new_version():
        from GangaCore.Utility.Config.Config import getConfig
        versions_filename = os.path.join(os.path.expanduser(os.path.expandvars(getConfig('Configuration')['used_versions_path'])), '.used_versions')
        if not os.path.exists(os.path.dirname(versions_filename)):
            os.makedirs(os.path.dirname(versions_filename))
        if not os.path.exists(versions_filename):
            with open(versions_filename, 'w') as versions_file:
                versions_file.write(_gangaVersion + '\n')
            return True, True

        majVersion, minVersion, bugVersion = _gangaVersion.split('.')
        newMinVersion = False
        minVersion = majVersion+'.'+minVersion+'.'
        # Check to see if this is a new bugfix version
        with open(versions_filename, 'r+') as versions_file:
            if versions_file.read().find(_gangaVersion) < 0:
                # Check to see if this is a new minor version as well
                versions_file.seek(0)
                if versions_file.read().find(minVersion) < 0:
                    newMinVersion = True
                versions_file.write(_gangaVersion + '\n')
                return True, newMinVersion

        return False, newMinVersion

    @staticmethod
    def rollHistoryForward():
        from GangaCore.Utility.Config.Config import getConfig
        versions_filename = os.path.join(os.path.expanduser(os.path.expandvars(getConfig('Configuration')['used_versions_path'])), '.used_versions')
        hasLoaded_newer = False
        with open(versions_filename, 'r') as versions_file:
            this_version = versions_file.read()
            split_version_info = this_version.split('.')
            if len(split_version_info) == 3:
                major = int(split_version_info[0])
                minor = int(split_version_info[1])
                debug = int(split_version_info[2])
                if major > 6:
                    hasLoaded_newer = True
                if major == 6 and minor > 1:
                    hasLoaded_newer = True
                if major == 6 and minor == 1 and debug > 13:
                    hasLoaded_newer = True
        
        old_dir = os.path.expanduser('~/.ipython-ganga')
        if 'IPYTHONDIR' in os.environ:
            old_dir = os.path.abspath(os.path.expanduser(os.environ['IPYTHONDIR']))

        single_pass_file = os.path.join(old_dir, '.have_migrated')
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("testing: %s" % single_pass_file)

        if not os.path.exists(single_pass_file) and os.path.exists(os.path.join(old_dir, 'history')):

            logger.info("This is your first time running Ganga >=6.1.14")
            logger.info("Now attempting to migrate your IPython history file")

            try:
                from IPython.core.interactiveshell import InteractiveShell
                from IPython.core.history import HistoryManager
                from IPython.utils.path import locate_profile, get_ipython_dir
            except ImportError as err:
                logger.error("Unable to import required IPython files, can't roll forward history")
                logger.error("If you want to roll forward your old history files you'll need to do this by hand")
                logger.error("Please try running: https://gist.github.com/minrk/6003365")

            old_text_history = os.path.join(old_dir, 'history')
            new_sqlite_history = os.path.join(old_dir, "profile_default/history.sqlite")

            if not os.path.exists(os.path.join(old_dir, "profile_default")):
                os.makedirs(os.path.join(old_dir, "profile_default"))

            IPython_history = HistoryManager(hist_file=new_sqlite_history)#, shell = InteractiveShell.instance())

            with open(old_text_history) as hist_file:
                for linenum, line in enumerate(hist_file):
                    IPython_history.store_inputs(linenum, line)

            ## Set a file to indicate that we've already made this transition
            passed_file = open(single_pass_file, "w")
            passed_file.close()

            logger.info("IPython history migrated successfully.")


    @staticmethod
    def generate_config_file(config_file, interactive):
        from GangaCore.GPIDev.Lib.Config.Config import config_file_as_text
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()

        # Old backup routine
        if os.path.exists(config_file):
            for i in range(100):
                bn = "%s.%.2d" % (config_file, i)
                if not os.path.exists(bn):
                    try:
                        os.rename(config_file, bn)
                    except Exception, err:
                        logger.error('Failed to create config backup file %s' % bn)
                        logger.error('Old file will not be overwritten, please manually remove it and start ganga with the -g option to re-generate it')
                        logger.error('Reason: %s' % err)
                        return
                    logger.info('Copied current config file to %s' % bn)
                    break
            else:
                config_directory = os.path.dirname(os.path.abspath(config_file))
                config_backupdir = os.path.join(config_directory, '.gangarc_backups')
                if not os.path.exists(config_backupdir):
                    os.makedirs(config_backupdir)

                datestr = "_" + time.strftime("%d.%m.%y")
                logger.info('Copying backup config files to %s' %
                            config_backupdir)
                for i in range(100):
                    old_bn = "%s.%.2d" % (config_file, i)
                    bn = os.path.basename(old_bn)
                    new_bn = bn + datestr
                    new_bn_file = os.path.join(config_backupdir, new_bn)
                    os.rename(old_bn, new_bn_file)
                bn = "%s.%.2d" % (config_file, 0)
                os.rename(config_file, bn)
                logger.info('Copied current config file to %s' % bn)
                #raise ValueError('too many bckup files')

        logger.info('Creating ganga config file %s' % config_file)
        new_config = ''

        with open(os.path.join(os.path.dirname(GangaCore.Runtime.__file__), 'HEAD_CONFIG.INI'), 'r') as config_head_file:
            new_config += config_head_file.read()
        new_config += config_file_as_text(interactive)
        new_config = new_config.replace('Ganga-SVN', new_version_format_to_old(_gangaVersion))  #Add in Ganga-x-y-z format so that it is backward compatible.
        with open(config_file, 'w') as new_config_file:
            new_config_file.write(new_config)

    @staticmethod
    def print_release_notes():
        from GangaCore.Utility.logging import getLogger
        from GangaCore.Utility.Config.Config import getConfig
        import itertools
        logger = getLogger('ReleaseNotes')
        if getConfig('Configuration')['ReleaseNotes'] == True:
            packages = itertools.imap(lambda x: 'ganga/ganga/' + x, itertools.ifilter(lambda x: x != '', ['GangaCore'] + getConfig('Configuration')['RUNTIME_PATH'].split(':')))
            pathname = os.path.join(os.path.dirname(__file__), '..', '..', 'GangaRelease', 'ReleaseNotes-%s' % _gangaVersion)

            if not os.path.exists(pathname):
                logger.warning("couldn't find release notes for version %s" % _gangaVersion)
                return

            bounding_line = '**************************************************************************************************************\n'
            dividing_line = '--------------------------------------------------------------------------------------------------------------\n'
            with open(pathname, 'r') as f:
                try:
                    notes = [l.strip() for l in f.read().replace(bounding_line, '').split(dividing_line)]
                except Exception, err:
                    logger.error('Error while attempting to read release notes')
                    logger.debug('Reason: %s' % err)
                    raise

            if notes[0].find(_gangaVersion) < 0:
                logger.error("Release notes version doesn't match the stated version on line 1")
                logger.error("'%s' does not match '%s'" % (_gangaVersion, notes[0]))
                return

            log_divider = '-' * 50
            note_gen = [(p, notes[notes.index(p) + 1].splitlines()) for p in packages if p in notes]
            if note_gen:
                logger.info(log_divider)
                logger.info(log_divider)
                logger.info("Release notes for version 'Ganga-%s':" % _gangaVersion)
                logger.info(log_divider)
                logger.info('')
                logger.info(log_divider)
                for p, n in note_gen:
                    logger.info(p)
                    logger.info(log_divider)
                    # logger.info('*'*len(p))
                    logger.info('')
                    for l in n:
                        logger.info(l.strip())
                    logger.info('')
                    logger.info(log_divider)
                logger.info(log_divider)

    # this is an option method which runs an interactive wizard which helps new users to start with Ganga
    # the interactive mode is not entered if -c option was used
    def new_user_wizard(self, interactive=True):
        from GangaCore.Utility.logging import getLogger
        from GangaCore.Utility.Config.Config import load_user_config, getConfig, ConfigError

        logger = getLogger('NewUserWizard')
        specified_gangadir = os.path.expanduser(os.path.expandvars(getConfig('Configuration')['gangadir']))
        specified_config = self.options.config_file
        default_gangadir = os.path.expanduser('~/gangadir')
        default_config = self.default_config_file

        if not os.path.exists(specified_gangadir) \
                and not os.path.exists(specified_config):
            logger.info('It seems that you run Ganga for the first time')
            logger.info('Ganga will send a udp packet each time you start it in order to help the development team understand how ganga is used. You can disable this in the config file by resetting [Configuration]UsageMonitoringURL=  ')

        if not os.path.exists(specified_gangadir) \
                and not os.path.exists(default_gangadir):
            logger.info('Making default gangadir: %s' % gangadir)
            try:
                os.makedirs(gangadir)
            except OSError as err:
                logger.error("Failed to create default gangadir '%s': %s" % (gangadir, err.message))
                raise
        if self.options.generate_config:
            logger = getLogger('ConfigUpdater')
            logger.info('re-reading in old config for updating...')
            load_user_config(specified_config, {})
            self.generate_config_file(specified_config, interactive)
            sys.exit(0)
        bugVer, minVer = self.new_version()
        if not os.path.exists(specified_config) \
                and not os.path.exists(default_config):
            # Sleep for 1 sec to allow for most of the bootstrap to finish so
            # the user actually sees this message last
            time.sleep(3.)
            if interactive:
                yes = raw_input('Would you like to create default config file ~/.gangarc with standard settings ([y]/n) ?\n')
            else:
                yes = 'y'
            if yes.lower() in ['', 'y']:
                self.generate_config_file(default_config, interactive)
                if interactive:
                    raw_input('Press <Enter> to continue.\n')
        elif bugVer:
            self.print_release_notes()
            self.rollHistoryForward()
            # if config explicitly set we dont want to update the versions file
            # this is so that next time ganga used with the default .gangarc
            # it will still be classed as new so the default config is updated.
            # also if config explicitly set dont try to update it for new version
            # impacts hammercloud if you do?
            if not self.options.config_file_set_explicitly:
                if minVer:
                    logger = getLogger('ConfigUpdater')
                    logger.info('It appears that this is the first time you have run %s' % _gangaVersion)
                    logger.info('Your ganga config file will be updated.')
                    logger.info('re-reading in old config for updating...')
                    load_user_config(specified_config, {})
                    self.generate_config_file(specified_config, interactive)

                # config file generation overwrites user values so we need to reapply the cmd line options to these user settings
                # e.g. set -o[Configuration]gangadir=/home/mws/mygangadir and the user value gets reset to the .gangarc value
                # (not the session value but the user value has precedence)
                try:
                    opts = self.parse_cmdline_config_options(self.options.cmdline_options)
                    for section, option, val in opts:
                        config = getConfig(section).setUserValue(option, val)
                except ConfigError as x:
                    self.exit('command line option error when resetting after config generation: %s' % x)

    @staticmethod
    def parse_cmdline_config_options(cmdline_options):
        """ Parse a list of command line config options and return a list of triplets (section,option,value).
        In case of parsing errors, raise ConfigError exception.
        """
        mpat = re.compile(r'(\[(?P<section>\S+)\]|)(?P<option>[a-zA-z0-9._/]+)=(?P<value>.+)')
        section = None

        opts = []
        for o in cmdline_options:
            rpat = mpat.match(o)
            if rpat is None:
                from GangaCore.Utility.Config import ConfigError
                raise ConfigError('syntax error: "%s"' % o)
            else:
                if rpat.group('section'):
                    section = rpat.group('section')
                if section is None:
                    from GangaCore.Utility.Config import ConfigError
                    raise ConfigError('section not specified: %s' % o)
                else:
                    opts.append((section, rpat.group('option'), rpat.group('value')))
        return opts

    @staticmethod
    def get_config_files(config_file, config_path=None):
        # type: (str, str) -> List[str]
        """
        Get the list of the config files to read. They are in order
        of increasing precedence so those later in the list have a
        higher priority.

        Args:
            config_file (str): the user config file
            config_path (str): the path to the config directory

        Returns:
            a list of strings of the filenames of the config files
        """
        from GangaCore.Utility.logging import getLogger

        try:
            with open(config_file) as cf:
                first_line = cf.readline()
                r = re.compile('# Ganga configuration file \(\$[N]ame: (?P<version>\S+) \$\)').match(first_line)
                this_logger = getLogger("Configure")
                if not r:
                    this_logger.error('file %s does not seem to be a Ganga config file', config_file)
                    this_logger.error('try -g option to create valid ~/.gangarc')
                else:
                    cv = r.group('version').split('-')  #Version number is in Ganga-x-y-z format
                    if len(cv) == 1:
                        cv = new_version_format_to_old(cv[0]).split('-')
                    if cv[1] != '7':
                        this_logger.error('file %s was created by a development release (%s)', config_file, r.group('version'))
                        this_logger.error('try -g option to create valid ~/.gangarc')
        except IOError as x:
            # ignore all I/O errors (e.g. file does not exist), this is just an
            # advisory check
            this_logger = getLogger("Configure")
            this_logger.debug("Config File Exception: %s" % x)

        if config_path is None:
            try:
                config_path = os.environ['GANGA_CONFIG_PATH']
            except KeyError:
                config_path = ''
            if config_path is None:
                config_path = ''

        import GangaCore.Utility.files
        import GangaCore.Utility.util
        import inspect
        GangaRootPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), '../..'))

        if config_path:
            config_path = GangaCore.Utility.files.expandfilename(os.path.join(GangaRootPath, config_path))

        # check if the specified config options are different from the defaults
        # and set session values appropriately
        syscfg = getConfig("System")
        if config_path != syscfg['GANGA_CONFIG_PATH']:
            syscfg.setSessionValue('GANGA_CONFIG_PATH', config_path)
        if config_file != syscfg['GANGA_CONFIG_FILE']:
            syscfg.setSessionValue('GANGA_CONFIG_FILE', config_file)

        # all relative names in the path are resolved wrt the _gangaPythonPath
        # the list order is reversed so that A:B maintains the typical path precedence: A overrides B
        # because the user config file is put at the end it always may override
        # everything else
        config_files = GangaCore.Utility.Config.expandConfigPath(config_path, _gangaPythonPath)
        config_files.reverse()

        def _createpath(dir):

            def _accept(fname, p=re.compile('.*\.ini$')):
                return (os.path.isfile(fname) or os.path.islink(fname)) and p.match(fname)

            files = []
            if dir and os.path.exists(dir) and os.path.isdir(dir):
                files = [os.path.join(dir, f) for f in os.listdir(dir) if
                         _accept(os.path.join(dir, f))]
            return string.join(files, os.pathsep)

        def _versionsort(s, p=re.compile(r'^(\d+)-(\d+)-*(\d*)')):
            m = p.match(s)
            if m:
                if m.group(3) == '':
                    return int(m.group(1)), int(m.group(2)), 0
                else:
                    return int(m.group(1)), int(m.group(2)), int(m.group(3))
            if s == 'SVN':
                return 'SVN'
            return None

        if "GANGA_SITE_CONFIG_AREA" in os.environ:
            this_dir = os.environ['GANGA_SITE_CONFIG_AREA']
            if os.path.exists(this_dir) and os.path.isdir(this_dir):
                dirlist = sorted(os.listdir(this_dir), key=_versionsort)
                dirlist.reverse()
                gangaver = _versionsort(new_version_format_to_old(_gangaVersion).lstrip('Ganga-'))  # Site config system expects x-y-z version encoding
                for d in dirlist:
                    vsort = _versionsort(d)
                    if vsort and ((vsort <= gangaver) or (gangaver is 'SVN')):
                        select = os.path.join(this_dir, d)
                        config_files.append(_createpath(select))
                        break
        if os.path.exists(config_file):
            config_files.append(config_file)

        return config_files

    # configuration procedure: read the configuration files, configure and
    # bootstrap logging subsystem
    def configure(self, logLevel=None):

        import GangaCore.Utility.Config
        from GangaCore.Utility.Config import ConfigError

        def set_cmdline_config_options(sects=None):
            try:
                opts = self.parse_cmdline_config_options(self.options.cmdline_options)
                for section, option, val in opts:
                    should_set = True
                    if not sects is None and not section in sects:
                        should_set = False
                    if should_set:
                        config = GangaCore.Utility.Config.setUserValue(section, option, val)
            except ConfigError as x:
                self.exit('command line option error: %s' % x)

        # set logging options
        set_cmdline_config_options(sects=['Logging'])

        # Say Hello
        if logLevel:
            self.options.force_loglevel = logLevel

        if self.options.force_loglevel in (None, 'DEBUG'):
            sys.stdout.write(str(self.hello_string) + '\n')

        if self.options.config_file is None or self.options.config_file == '':
            self.options.config_file = self.default_config_file

        # initialize logging for the initial phase of the bootstrap
        # will use the default, hardcoded log level in the module until
        # pre-configuration procedure is complete
        from GangaCore.Utility.logging import force_global_level, getLogger

        force_global_level(self.options.force_loglevel)

        # Assemble the list of config files to read
        config_files = self.get_config_files(self.options.config_file)

        syscfg = getConfig("System")

        # read-in config files

        # FIXME: need to construct a proper dictionary - cannot use the
        # ConfigPackage directly
        system_vars = {}
        for opt in syscfg:
            system_vars[opt] = syscfg[opt]

        def deny_modification(name, x):
            raise GangaCore.Utility.Config.ConfigError(
                'Cannot modify [System] settings (attempted %s=%s)' % (name, x))
        syscfg.attachUserHandler(deny_modification, None)
        syscfg.attachSessionHandler(deny_modification, None)
        GangaCore.Utility.Config.setSessionValuesFromFiles(config_files, system_vars)

        # set the system variables to the [System] module
        # syscfg.setDefaultOptions(system_vars,reset=1)

        # activate the logging subsystem
        # user defined log level takes effect NOW
        from GangaCore.Utility.logging import bootstrap
        bootstrap()

        if not self.options.monitoring:
            self.options.cmdline_options.append('[PollThread]autostart=False')
            self.options.cmdline_options.append('[Tasks]disableTaskMon=True')

        logger = getLogger()

        logger.debug('user specified cmdline_options: %s', self.options.cmdline_options)

        # override the config options from the command line arguments
        # the format is [section]option=value OR option=value
        # in the second case last specified section from previous options is
        # used

        set_cmdline_config_options()

        if self.options.TEST:
            # FIXME: CONFIG CHECK
            # ?? config['RUNTIME_PATH'] = ''
            config = getConfig('Configuration')
            exiting = config['RUNTIME_PATH']
            config.setSessionValue('RUNTIME_PATH', exiting + ':GangaTest')

        # ensure we're not interactive if daemonised
        if self.options.daemon and self.interactive:
            logger.warning("Cannot run as a service in interactive mode. Ignoring.")
            self.options.daemon = False

        # fork ourselves so we a daemon
        if self.options.daemon:
            logger.debug("Daemonising GangaCore...")
            pid = os.fork()
            if pid < 0:
                sys.exit(1)

            if pid > 0:
                sys.exit(0)

            os.umask(0)
            os.setsid()
            pid = os.fork()

            if pid > 0:
                sys.exit(0)

            # change the stdout/err
            sys.stdout.flush()
            sys.stderr.flush()

            config = getConfig("Configuration")

            # create a server dir
            if not os.path.exists(os.path.join(config['gangadir'], "server")):
                os.makedirs(os.path.join(config['gangadir'], "server"))

            si = file("/dev/null", 'r')
            so = file(os.path.join(
                config['gangadir'], "server", "server-%s.stdout" % (os.uname()[1])), 'a')
            se = file(os.path.join(
                config['gangadir'], "server", "server-%s.stderr" % (os.uname()[1])), 'a', 0)

            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())

    # initialize environment: find all user-defined runtime modules and set their environments
    @staticmethod
    def initEnvironment():

        from GangaCore.Utility.logging import getLogger
        logger = getLogger()

        logger.debug("Installing Shutdown Manager")
        from GangaCore.Core.InternalServices.ShutdownManager import register_exitfunc
        register_exitfunc()

        import GangaCore.Utility.Config
        from GangaCore.Utility.Runtime import initSetupRuntimePackages
        from GangaCore.Core.exceptions import GangaException

        logger.debug("Import plugins")
        try:
            # load Ganga system plugins...
            from GangaCore.Runtime import plugins
        except Exception as x:
            logger.critical('Ganga system plugins could not be loaded due to the following reason: %s', x)
            logger.exception(x)
            raise GangaException(x), None, sys.exc_info()[2]

        initSetupRuntimePackages()

        from GangaCore.Core.GangaThread.WorkerThreads import startUpQueues
        startUpQueues()

    # bootstrap all system and user-defined runtime modules
    @staticmethod
    def bootstrap(interactive):
        import GangaCore.Utility.Config
        config = GangaCore.Utility.Config.getConfig('Configuration')

        from GangaCore.Utility.Runtime import loadPlugins, autoPopulateGPI
        loadPlugins(GangaCore.GPI)
        autoPopulateGPI()
        from GangaCore.Utility.Runtime import setPluginDefaults
        setPluginDefaults()

        # Start tracking all the threads and saving the information to a file
        stacktracer.trace_start()

        from GangaCore.Utility.logging import getLogger
        logger = getLogger()

        manualExportToGPI()

        from GangaCore.Runtime import Workspace_runtime, Repository_runtime
        from GangaCore.GPIDev.Credentials import credential_store
        if (Workspace_runtime.requiresAfsToken() or Repository_runtime.requiresAfsToken()) and not config['NoAfsToken']:
            # If the registry or the workspace needs an AFS token then add one to the credential store.
            # Note that this happens before the monitoring starts so that it gets tracked properly

            from GangaCore.GPIDev.Credentials.AfsToken import AfsToken
            credential_store.create(AfsToken(), create=False, check_file=True)

        import GangaCore.Core
        from GangaCore.Runtime.Repository_runtime import startUpRegistries
        if config['AutoStartReg']:
            startUpRegistries()

        logger.debug("Bootstrap Core Modules")
        # bootstrap core modules
        from GangaCore.Core.GangaRepository import getRegistrySlice
        ## Here is where the monitoring loop and related services are started!
        GangaCore.Core.bootstrap(getRegistrySlice('jobs'), interactive)

        # export all configuration items, new options should not be added after
        # this point
        GangaCore.GPIDev.Lib.Config.bootstrap()


        logger.debug("Post-Bootstrap hooks")
        from GangaCore.Utility.Runtime import allRuntimes
        ###########
        # run post bootstrap hooks
        for r in allRuntimes.values():
            try:
                r.postBootstrapHook()
            except Exception as err:
                logger.error("problems with post bootstrap hook for %s" % r.name)
                logger.error("Reason: %s" % err)


    @staticmethod
    def startTestRunner(my_args):
        """
        run the testing framework
        """

        from GangaCore.Utility.logging import getLogger
        logger = getLogger()

        try:
            # Important to avoid a lot of arguments over who has locked what object,
            # the tests are quite intensive and often trip on the background
            # monitoring thread(s)
            from GangaCore.Core.InternalServices.Coordinator import disableMonitoringService
            disableMonitoringService()

            from GangaTest.Framework import runner
            from GangaTest.Framework import htmlizer
            from GangaTest.Framework import xmldifferencer

            tfconfig = GangaCore.Utility.Config.getConfig('TestingFramework')
            rc = 1
            if tfconfig['EnableTestRunner']:
                logger.info("Starting Ganga Test Runner")

                if not my_args:
                    logger.warning("Please specify the tests to run ( i.e. ganga --test Ganga/old_test )")
                    return -1

                logger.info("myargs = %s" % " ".join(my_args))

                rc = runner.start(test_selection=" ".join(my_args))
            else:
                logger.info("Test Runner is disabled (set EnableTestRunner=True to enable it)")

            if rc > 0 and tfconfig['EnableHTMLReporter']:
                logger.info("Generating tests HTML reports")
                rc = htmlizer.main(tfconfig)
            elif rc > 0 and tfconfig['EnableXMLDifferencer']:
                logger.info("Generating difference HTML reports")
                rc = xmldifferencer.main(my_args)
            return rc
        except ImportError as e:
            logger.error("You need GangaTest external package in order to invoke Ganga test-runner.")
            logger.error(e)
            return -1

    # run Ganga in the specified namespace, in principle the namespace should import all names from GangaCore.GPI
    # if namespace is not specified then run in __main__
    def run(self, local_ns=None):

        from GangaCore.Utility.logging import getLogger
        logger = getLogger("run")
        logger.debug("Entering run")

        if self.options.webgui == True:
            from GangaCore.Runtime.http_server import start_server
            start_server()

        if local_ns is None:
            import __main__
            local_ns = __main__.__dict__
        # save a reference to the Ganga namespace as an instance attribute
        self.local_ns = local_ns

        # load templates for user-defined runtime modules
        from GangaCore.Utility.Runtime import allRuntimes
        for r in allRuntimes.values():
            r.loadTemplates(local_ns)

        # exec ~/.ganga.py file
        fileName = fullpath('~/.ganga.py')
        if os.path.exists(fileName):
            try:
                execfile(fileName, local_ns)
            except Exception as x:
                logger.error('Failed to source %s (Error was "%s"). Check your file for syntax errors.', fileName, x)
        # exec StartupGPI code
        from GangaCore.Utility.Config import getConfig
        config = getConfig('Configuration')
        if config['StartupGPI']:
            # ConfigParser trims the lines and escape the space chars
            # so we have only one possibility to insert python code :
            # using explicitly '\n' and '\t' chars
            code = config['StartupGPI'].replace(
                '\\t', '\t').replace('\\n', '\n')
            exec code in local_ns

        logger.debug("loaded .ganga.py")

        # this logic is a bit convoluted
        runs_script = len(self.args) > 0
        session_type = config['TextShell']
        if runs_script:
            if not self.interactive:
                session_type = 'batch'
            else:
                session_type += 'startup_script'

        if self.options.TEST:
            sys.argv = self.args
            try:
                rc = self.startTestRunner(self.args)
            except (KeyboardInterrupt, SystemExit):
                logger.warning('Test Runner interrupted!')
                import GangaCore.Core.InternalServices.Coordinator
                if not GangaCore.Core.InternalServices.Coordinator.servicesEnabled:
                    from GangaCore.Core.InternalServices.Coordinator import enableInternalServices
                    enableInternalServices()
                sys.exit(1)
            sys.exit(rc)

        if len(self.args) > 0:
            # run the script and make it believe it that it is running directly
            # as an executable (sys.argv)
            saved_argv = sys.argv
            sys.argv = self.args

            import GangaCore.Utility.Runtime
            path = GangaCore.Utility.Runtime.getSearchPath()
            script = GangaCore.Utility.Runtime.getScriptPath(self.args[0], path)

            #if script:
            #    script_file = open(script, 'r')
            #    script_content = script_file.read()
            #    script_file.close()
            #
            #    compiled_script = compile(script_content, '<string>', 'exec')
            #
            #exec compiled_script

            if script:
                from GangaCore.Utility.Config.Config import setConfigOption
                if not getConfig('PollThread').hasModified:
                    setConfigOption('PollThread','forced_shutdown_policy', 'timeout')
                from GangaCore.Core.GangaThread import GangaThreadPool
                GangaThreadPool.shutdown_policy = 'batch'
                execfile(script, local_ns)
            else:
                logger.error("'%s' not found" % self.args[0])
                logger.info("Searched in path %s" % path)
                sys.exit(1)
            sys.argv = saved_argv

        # and exit unless -i was specified
        if not self.interactive:
            return
            #    sys.exit(0)

        # interactive python shell

        shell = config['TextShell']

        if shell == 'IPython':

            try:
                import IPython
            except ImportError as err:
                logger.error("Cannot load IPython class!!!")
                logger.error("Please check your environment and re-load Ganga from a working shell!")
                logger.error("Exiting Ganga now, goodbye!")
                return

            # buffering of log messages from all threads called "GANGA_Update_Thread"
            # the logs are displayed at the next IPython prompt

            from GangaCore.Utility.logging import enableCaching
            enableCaching()

            self.check_IPythonDir()
            self.launch_IPython(local_ns, GangaProgram._ganga_error_handler, self.ganga_prompt)

        else:
            override_credits()
            import code
            sys.displayhook = _display
            c = code.InteractiveConsole(locals=local_ns)
            c.interact()

        return

    @staticmethod
    def check_IPythonDir():
        """
        Check that the .ipython folder exists as expected as defined by IPYTHONDIR
        """

        from GangaCore.Utility.logging import getLogger
        logger = getLogger()

        not_exist = False

        if 'IPYTHONDIR' in os.environ:
            ipyth_dir = os.environ['IPYTHONDIR']
            os.environ['IPYTHONDIR'] = os.path.abspath(os.path.expanduser(os.environ['IPYTHONDIR']))
            #logger.warning('Environment variable IPYTHONDIR=%s exists and overrides the default history file for Ganga IPython commands', ipyth_dir)
            if not os.path.isdir(ipyth_dir):
                if os.path.exists(ipyth_dir):
                    ipyth_dir_bak = str(ipyth_dir).append('_bak')
                    logger.warning('IPYTHONDIR points to: %s which is not a directory, moving it to: %s' % (ipyth_dir, ipyth_dir_bak))
                    os.rename(ipyth_dir, ipyth_dir_bak)
                    not_exist = True
        else:
            newpath = os.path.expanduser('~/.ipython-ganga')
            oldpath = os.path.expanduser('~/.ipython')
            os.environ['IPYTHONDIR'] = newpath
            if not os.path.exists(newpath):
                if os.path.exists(oldpath):
                    logger.warning('Default location of IPython history files has changed.')
                    logger.warning('Ganga will now try to copy your old settings from %s to the new path %s.\
                            If you do not want that, quit Ganga and wipe off the content of new path: rm -rf %s/*', oldpath, newpath, newpath)
                    import shutil
                    shutil.copytree(oldpath, newpath)
                else:
                    not_exist = True

        if not_exist:
            os.makedirs(os.environ['IPYTHONDIR'])

        rc_file = os.path.join(os.environ['IPYTHONDIR'], 'ipythonrc')
	logger.debug("Checking: %s" % rc_file)
        if not os.path.isfile(rc_file):
            lock = open(rc_file, "w")
            lock.close()

        return None

    @staticmethod
    def _ganga_error_handler(exception_obj, etype, value, tb, tb_offset=None):
        """
        Error handler for IPython 3.x+ to identify expected Ganga exceptions or unexpected uncaught exceptions from somewhere
        """
        ## see https://ipython.org/ipython-doc/dev/api/generated/IPython.core.interactiveshell.html#IPython.core.interactiveshell.InteractiveShell.set_custom_exc
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()
        #logger.error("%s" % value)

        from GangaCore.Core.exceptions import GangaException
        import traceback
        # Extract the stack from this traceback object
        stack = traceback.extract_tb(tb)
        # If this is an error from the interactive prompt then the length is 2, otherwise the error is from deeper in Ganga
        if not issubclass(etype, GangaException) and len(stack) > 2:
            logger.error("!!Unknown/Unexpected ERROR!!")
            logger.error("If you're able to reproduce this please report this to the Ganga developers!")
            #logger.error("value: %s" % value)
            exception_obj.showtraceback((etype, value, tb), tb_offset=tb_offset)
        else:
            logger.error("%s" % value)
        return None

    @staticmethod
    def launch_IPython(local_ns, error_handler, ganga_prompt):
        """
        Launch an embedded IPython session within the GangaCore.GPI namespace
        """

        import IPython

        # Based on examples/Embedding/embed_class_long.py from the IPython source tree

        # First we set up the prompt
        from traitlets.config.loader import Config
        
        from GangaCore.Utility.Config import getConfig
        gangaconfig = getConfig('TextShell_IPython')
        cfg = Config()
        cfg.TerminalInteractiveShell.colors = gangaconfig['colourscheme']
        cfg.TerminalInteractiveShell.autocall = 0
        cfg.PlainTextFormatter.pprint = True
        banner = exit_msg = ''
        cfg.TerminalInteractiveShell.prompts_class = GangaPrompt

        # Import the embed function
        from IPython.terminal.embed import InteractiveShellEmbed

        ipshell = InteractiveShellEmbed(config=cfg, banner1=banner, exit_msg=exit_msg)
        ipshell.events.register("post_execute", ganga_prompt)

        # Add our custom error handler to ignore stack traces for GangaExceptions
        ipshell.set_custom_exc((Exception,), error_handler)

        # buffering of log messages from all threads called "GANGA_Update_Thread"
        # the logs are displayed at the next IPython prompt
        from GangaCore.Utility.logging import enableCaching
        enableCaching()

        # Magic ganga function for replacing 'execfile'
        from GangaCore.Runtime.IPythonMagic import GangaMagics
        ipshell.register_magics(GangaMagics)

        # Setting up the confirmation of exit on Ctrl+D this prompt annoys some users(developers)
        from GangaCore.Utility.Config.Config import getConfig
        config = getConfig('Configuration')
        ipshell.confirm_exit = config['confirm_exit']

        # Launch embedded shell
        from GangaCore import GPI
        ipshell(local_ns=local_ns, module=GangaCore.GPI)


    @staticmethod
    def ganga_prompt(_=None):

        try:
            from GangaCore.GPIDev.Credentials import get_needed_credentials

            needed_credentials = get_needed_credentials()

            # Add still-needed credentials to the prompt
            if needed_credentials:
                prompt = 'Warning, some credentials needed by the monitoring are missing or invalid:\n'
                for cred_req in needed_credentials:
                    prompt += '  ' + str(cred_req).replace('\n ', '') + '\n'
                prompt += 'Call `credential_store.renew()` to update them.\n'
                print(prompt)
        except KeyboardInterrupt:
            return


#
# $Log: not supported by cvs2svn $
# Revision 1.11.4.1  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.11  2009/04/28 13:37:12  kubam
# simplified handling of logging filters
#
# Revision 1.15  2009/07/20 14:13:44  moscicki
# workaround for wierd OSX execv behaviour (from Ole Weidner)
#
# Revision 1.14  2009/06/10 14:53:05  moscicki
# fixed bug #51592: Add self to logger
#
# Revision 1.13  2009/06/09 10:44:55  moscicki
# removed obsolete variable
#
# Revision 1.12  2009/06/08 15:48:17  moscicki
# fix Ganga to work with newer versions of ipython (-noautocall option was removed in newer ipython versions)
#
# Revision 1.11  2009/04/28 13:37:12  kubam
# simplified handling of logging filters
#
# Revision 1.10  2009/02/02 13:43:26  moscicki
# fixed: bug #44934: Didn't create .gangarc on first usage
#
# Revision 1.9  2008/11/27 15:49:03  moscicki
# extra exception output if cannot load the plugins...
#
# Revision 1.8  2008/11/21 16:34:22  moscicki
# bug #43917: Implement Batch backend as alias to default backend at a given site
#
# Revision 1.7  2008/10/23 15:24:04  moscicki
# install the shutdown manager for atexit handlers before loading system plugins (e.g. LCG download thread registers the atexit handler using a tuple (priority,handler))
#
# Revision 1.6  2008/09/05 15:55:51  moscicki
# XML differenciater added (from Ulrik)
#
# Revision 1.5  2008/08/18 13:18:59  moscicki
# added force_status() method to replace job.fail(), force_job_failed() and
# force_job_completed()
#
# Revision 1.4  2008/08/18 10:02:15  moscicki
#
# bugfix 40110
#
# Revision 1.3  2008/08/01 15:25:30  moscicki
# typo fix
#
# Revision 1.2  2008/07/31 17:25:02  moscicki
# config templates are now in a separate directory at top level ("templates")
#
# *converted all tabs to spaces*
#
# Revision 1.1  2008/07/17 16:41:00  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.71.4.23  2008/07/03 16:11:31  moscicki
# bug #38000: Add check for old .gangarc file
#
# Revision 1.71.4.22  2008/04/03 12:55:45  kuba
# importing core plugins before initEnvironment(), this fixes
# bug #35146: GangaAtlas is not starting due to gridshell call in __init__.py
#
# Revision 1.71.4.21  2008/04/01 14:08:03  roma
# automatic config file template generation (Vladimir)
#
# Revision 1.71.4.20  2008/03/31 15:32:46  kubam
# use more flexible logic for hidden classes
#
# Revision 1.71.4.19  2008/03/12 17:33:32  moscicki
# workaround for broken logging system: GangaProgram.log writes directly to stderr
#
# Revision 1.71.4.18  2008/03/11 15:24:51  moscicki
# merge from Ganga-5-0-restructure-config-branch
#
# Revision 1.71.4.17.2.1  2008/03/07 13:34:38  moscicki
# workspace component
#
# Revision 1.71.4.17  2008/03/06 14:14:55  moscicki
# streamlined session options logic
# moved sanityCheck to config.bootstrap()
#
# Revision 1.71.4.16  2008/03/06 11:26:52  amuraru
# fixed .ganga.py sourcing file
#
# Revision 1.71.4.15  2008/03/05 14:53:33  amuraru
# execute ~/.ganga.py file before executing StartupGPI code
#
# Revision 1.71.4.14  2008/02/28 10:08:32  amuraru
# *** empty log message ***
#
# Revision 1.71.4.13  2008/02/21 12:11:02  amuraru
# added [Shell] section to configure internal Shell utility
#
# Revision 1.71.4.12  2008/02/06 17:04:11  moscicki
# initialize external monitoring services subsystem
#
# Revision 1.71.4.11  2007/12/18 16:51:28  moscicki
# merged from XML repository branch
#
# Revision 1.71.4.10  2007/12/18 13:05:19  amuraru
# removed coverage code from boostrap (moved in GangaTest/Framework/driver.py)
#
# Revision 1.71.4.9  2007/12/13 16:33:02  moscicki
# export more GPI exceptions
#
# Revision 1.71.4.8  2007/12/10 18:55:55  amuraru
# merged changes from Ganga 4.4.4
#
# Revision 1.71.4.7  2007/11/14 11:41:46  amuraru
# 5.0 configuration updated
#
# Revision 1.71.4.6.2.1  2007/11/13 16:26:05  moscicki
# removed obsolete migration GPI commands
#
# Revision 1.71.4.6  2007/11/08 13:21:00  amuraru
# moved testconfig option defintion to GangaTest
#
# Revision 1.71.4.5  2007/11/07 15:10:04  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
#
# Revision 1.71.4.4  2007/11/02 15:20:32  moscicki
# moved addOption() before config bootstrap
#
# Revision 1.71.4.3  2007/10/31 13:39:45  amuraru
# update to the new config system
#
# Revision 1.71.4.2  2007/10/25 11:43:11  roma
# Config update
#
# Revision 1.71.4.1  2007/10/12 13:56:26  moscicki
# merged with the new configuration subsystem
#
# Revision 1.71.6.2  2007/10/09 07:31:56  roma
# Migration to new Config
#
# Revision 1.71.6.1  2007/09/25 09:45:12  moscicki
# merged from old config branch
#
# Revision 1.71.8.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.76  2007/11/26 12:13:27  amuraru
# decode tab and newline characters in StartupGPI option
#
# Revision 1.75  2007/11/05 12:33:54  amuraru
# fix bug #30891
#
# Revision 1.74  2007/10/29 14:04:08  amuraru
#  - added free disk space checking in [PollThread] configuration template
#  - added an extra check not to attempt the shutdown of the repository if this has already been stopped
#  - save the Ganga namespace as an attribute in GangaCore.Runtime._prog
#
# Revision 1.73  2007/10/10 14:47:46  moscicki
# updated doc-strings
#
# Revision 1.72  2007/09/25 15:12:04  amuraru
#
# usa GANGA_CONFIG_FILE  environment variable to set the user config file
#
# Revision 1.71  2007/09/11 16:54:52  amuraru
# catch the TestRunner KeyboardInterrupt
#
# Revision 1.70  2007/09/11 14:28:29  amuraru
# implemented FR #28406 to allow definition of GPI statements to be executed at
# startup
#
# Revision 1.69  2007/08/27 10:47:30  moscicki
# overriden credits() and copyright() (request #21906)
#
# Revision 1.68  2007/08/22 15:58:55  amuraru
# Runtime/bootstrap.py
#
# Revision 1.67  2007/08/14 14:47:01  amuraru
# automatically add GangaTest RT package when --test is used
#
# Revision 1.66  2007/08/13 17:22:27  amuraru
# - testing framework small fix
#
# Revision 1.65  2007/08/13 13:19:48  amuraru
# -added EnableTestRunner and EnableHTMLReported to control the testing framework in a more flexible way
# -added GANGA_CONFIG_FILE in [System] config
#
# Revision 1.64  2007/08/13 12:50:18  amuraru
# added EnableTestRunner and EnableHTMLReported to control the testing framework in a more flexible way
#
# Revision 1.63  2007/07/30 12:57:51  moscicki
# removing IPython autocall option (obsoletion of jobs[] syntax and putting jobs() as a replacement)
#
# Revision 1.62  2007/07/27 14:31:55  moscicki
# credential and clean shutdown updates from Adrian (from Ganga-4-4-0-dev-branch)
#
# Revision 1.61  2007/07/10 13:08:32  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.60  2007/06/07 10:25:02  amuraru
# bug-fix: guard against environment update for RuntimePackages exposing null environment dictionary
#
# Revision 1.59  2007/06/04 14:31:22  amuraru
# record start-time of ganga session
#
# Revision 1.58  2007/06/01 08:49:08  amuraru
# Disable the autocompletion of private attributes and methods starting with _ or __
#
# Revision 1.57  2007/05/21 16:07:57  amuraru
# integrated TestingFramework into ganga itsefl (ganga --test). [TestingFramework] section in Ganga config is used to control it.
# changed GangaCore.Runtime.bootstrap default log level to INFO
#
# Revision 1.56  2007/05/11 13:21:24  moscicki
# temporary functions to help getting jobs out of completing and submitting states
# force_job_completed(j): may be applied to completing jobs
# force_job_failed(j): may be applied to submitting or completing jobs
#
# Revision 1.55  2007/05/08 10:32:42  moscicki
# added short GPL license summary at startup and license() command in GPI for full print
#
# Revision 1.54.6.1  2007/06/18 07:44:57  moscicki
# config prototype
#
# Revision 1.54  2007/02/28 18:24:53  moscicki
# moved GangaException to GangaCore.Core
#
# Revision 1.53  2007/02/22 13:43:19  moscicki
# pass interactive flag to Core.bootstrap
#
# Revision 1.52  2007/01/25 15:52:39  moscicki
# mergefrom_Ganga-4-2-2-bugfix-branch_25Jan07 (GangaBase-4-14)
#
# Revision 1.51.2.3  2006/12/15 17:12:37  kuba
# added spyware at startup
#
# Revision 1.51.2.2  2006/11/24 14:52:56  amuraru
# only available credentials (afs/gridproxy) are exported
#
# Revision 1.51.2.1  2006/11/24 14:22:05  amuraru
# added support for peek() function
#
# Revision 1.51  2006/10/23 10:59:41  moscicki
# initialize [Configuration]LOAD_PATH
#
# Revision 1.50  2006/10/16 12:53:13  moscicki
# fix the SCRIPTS_PATH mechanism: the . is always in the path and the session level updates are prepending to the default value... fix for bug #20332 overview: Ganga/scripts not included in SCRIPTS_PATH in Atlas.ini
#
# Revision 1.49  2006/10/04 18:16:48  moscicki
# fixed bug #20333 overview: hostname function of Ganga/Utility/util.py sometimes fails
#
# Revision 1.48  2006/09/27 16:38:31  moscicki
# changed AfsToken -> afsToken, GridProxy -> gridProxy and made them real GPI proxy objects
#
# Revision 1.47  2006/09/15 14:23:31  moscicki
# Greeting message goes to stderr (requested by UNOSAT to use Ganga in CGI scripts).
#
# Revision 1.46  2006/08/29 15:11:10  moscicki
# fixed #18084 Additonal global objects for splitters, mergers etc
#
# Revision 1.45  2006/08/29 12:51:57  moscicki
# exported GridProxy and AfsToken singleton objects to GPI
#
# Revision 1.44  2006/08/11 13:13:06  adim
# Added: GangaException as a markup base class for all exception that need to be printed in a usable way in IPython shell
#
# Revision 1.43  2006/08/09 09:07:34  moscicki
# added magic functions ('ganga')
#
# Revision 1.42  2006/07/31 12:13:43  moscicki
# depend on monitoring thread names "GANGA_Update_Thread" to do message buffering in IPython
#
# Revision 1.41  2006/07/27 20:21:24  moscicki
# - fixed option parsing
# - pretty formatting of known exceptions in IPython (A.Muraru)
#
# Revision 1.40  2006/06/21 11:43:00  moscicki
# minor fix
#
# Revision 1.39  2006/03/14 14:53:14  moscicki
# updated comments
#
# Revision 1.38  2006/03/09 08:41:52  moscicki
# --gui option and GUI integration
#
# Revision 1.37  2006/02/13 15:21:25  moscicki
# support for cached logging messages at interactive prompt (messages from monitoring thread are cached in IPython environment and printed at the next prompt)
#
# Revision 1.36  2006/02/10 14:16:00  moscicki
# fixed bugs:
# #13912        Cannot use tilde to give location of INI file
# #14436 problem with -o option at the command line and setting config default for properties
#
# exported ConfigError to GPI
# docstring updates
#
# Revision 1.35  2005/11/25 09:57:37  moscicki
# exported TreeError exception
#
# Revision 1.34  2005/11/14 14:47:38  moscicki
# jobtree added
#
# Revision 1.33  2005/11/14 10:29:10  moscicki
# support for default plugins
# temporary hack for GUI-specific monitoring
#
# Revision 1.32  2005/11/01 11:21:37  moscicki
# support for export/load (KH)
#
# Revision 1.31  2005/10/14 12:54:38  moscicki
# ignore I/O exceptions while checking for ganga3 config file
#
# Revision 1.30  2005/10/12 13:35:23  moscicki
# renamed _gangadir into gangadir
#
# Revision 1.29  2005/10/07 15:08:45  moscicki
# renamed __Ganga4__ into _gangadirm .ganga4 into .gangarc
# added sanity checks to detect old (Ganga3) config files
# added config-path mechanism
#
# Revision 1.28  2005/10/07 08:27:00  moscicki
# all configuration items have default values
#
# Revision 1.27  2005/09/22 12:48:14  moscicki
# import fix
#
# Revision 1.26  2005/09/21 09:12:50  moscicki
# added interactive displayhooks based on obj._display (if exists)
#
# Revision 1.25  2005/08/26 10:12:06  moscicki
# added [System] section (write-protected) with GANGA_VERSION and GANGA_PYTHONPATH (new)
#
# Revision 1.24  2005/08/24 15:24:11  moscicki
# added docstrings for GPI objects and an interactive ganga help system based on pydoc
#
#
#



"""Generic submit / finish / extract / report action implementations.

See CoreSubmitter, CoreFinisher, CoreExtractor and CoreReporter for core
implementations providing generic submit, finish, extract and report actions by
extending the base implementations in GangaRobot.Lib.Base.

Specific action implementations can reuse or extend these core implementations.

See GangaRobot.Lib.Example for implementations which extend these reuse these
core implementations.

"""


################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PACKAGE.py,v 1.9 2009-06-18 08:35:46 dvanders Exp $
################################################################################

""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""

_external_packages = {
}

from GangaCore.Utility.Setup import PackageSetup

setup = PackageSetup(_external_packages)

def standardSetup(setup=setup):
    for p in setup.packages:
        setup.prependPath(p,'LD_LIBRARY_PATH')
        setup.prependPath(p,'PATH')
        setup.setPath(p,'PANDA_SYS')

    # update the syspath and PYTHONPATH for the packages
    for name in setup.packages:
        setup.setSysPath(name)



#! /usr/bin/env python

"""Perform massive transformations on a document tree created from the LaTeX
of the Python documentation, and dump the ESIS data for the transformed tree.
"""


import errno
import esistools
import re
import sys
import xml.dom
import xml.dom.minidom

ELEMENT = xml.dom.Node.ELEMENT_NODE
ENTITY_REFERENCE = xml.dom.Node.ENTITY_REFERENCE_NODE
TEXT = xml.dom.Node.TEXT_NODE


class ConversionError(Exception):
    pass


ewrite = sys.stderr.write
try:
    # We can only do this trick on Unix (if tput is on $PATH)!
    if sys.platform != "posix" or not sys.stderr.isatty():
        raise ImportError
    import commands
except ImportError:
    bwrite = ewrite
else:
    def bwrite(s, BOLDON=commands.getoutput("tput bold"),
               BOLDOFF=commands.getoutput("tput sgr0")):
        ewrite("%s%s%s" % (BOLDON, s, BOLDOFF))


PARA_ELEMENT = "para"

DEBUG_PARA_FIXER = 0

if DEBUG_PARA_FIXER:
    def para_msg(s):
        ewrite("*** %s\n" % s)
else:
    def para_msg(s):
        pass


def get_first_element(doc, gi):
    for n in doc.childNodes:
        if n.nodeName == gi:
            return n

def extract_first_element(doc, gi):
    node = get_first_element(doc, gi)
    if node is not None:
        doc.removeChild(node)
    return node


def get_documentElement(node):
    result = None
    for child in node.childNodes:
        if child.nodeType == ELEMENT:
            result = child
    return result


def set_tagName(elem, gi):
    elem.nodeName = elem.tagName = gi


def find_all_elements(doc, gi):
    nodes = []
    if doc.nodeName == gi:
        nodes.append(doc)
    for child in doc.childNodes:
        if child.nodeType == ELEMENT:
            if child.tagName == gi:
                nodes.append(child)
            for node in child.getElementsByTagName(gi):
                nodes.append(node)
    return nodes

def find_all_child_elements(doc, gi):
    nodes = []
    for child in doc.childNodes:
        if child.nodeName == gi:
            nodes.append(child)
    return nodes


def find_all_elements_from_set(doc, gi_set):
    return __find_all_elements_from_set(doc, gi_set, [])

def __find_all_elements_from_set(doc, gi_set, nodes):
    if doc.nodeName in gi_set:
        nodes.append(doc)
    for child in doc.childNodes:
        if child.nodeType == ELEMENT:
            __find_all_elements_from_set(child, gi_set, nodes)
    return nodes


def simplify(doc, fragment):
    # Try to rationalize the document a bit, since these things are simply
    # not valid SGML/XML documents as they stand, and need a little work.
    documentclass = "document"
    inputs = []
    node = extract_first_element(fragment, "documentclass")
    if node is not None:
        documentclass = node.getAttribute("classname")
    node = extract_first_element(fragment, "title")
    if node is not None:
        inputs.append(node)
    # update the name of the root element
    node = get_first_element(fragment, "document")
    if node is not None:
        set_tagName(node, documentclass)
        # Move everything that comes before this node into this node;
        # this will be the document element.
        nodelist = fragment.childNodes
        point = node.firstChild
        while not nodelist[0].isSameNode(node):
            node.insertBefore(nodelist[0], point)
    while 1:
        node = extract_first_element(fragment, "input")
        if node is None:
            break
        inputs.append(node)
    if inputs:
        docelem = get_documentElement(fragment)
        inputs.reverse()
        for node in inputs:
            text = doc.createTextNode("\n")
            docelem.insertBefore(text, docelem.firstChild)
            docelem.insertBefore(node, text)
        docelem.insertBefore(doc.createTextNode("\n"), docelem.firstChild)
    while fragment.firstChild and fragment.firstChild.nodeType == TEXT:
        fragment.removeChild(fragment.firstChild)


def cleanup_root_text(doc):
    discards = []
    skip = 0
    for n in doc.childNodes:
        prevskip = skip
        skip = 0
        if n.nodeType == TEXT and not prevskip:
            discards.append(n)
        elif n.nodeName == "COMMENT":
            skip = 1
    for node in discards:
        doc.removeChild(node)


DESCRIPTOR_ELEMENTS = (
    "cfuncdesc", "cvardesc", "ctypedesc",
    "classdesc", "memberdesc", "memberdescni", "methoddesc", "methoddescni",
    "excdesc", "funcdesc", "funcdescni", "opcodedesc",
    "datadesc", "datadescni",
    )

def fixup_descriptors(doc, fragment):
    sections = find_all_elements(fragment, "section")
    for section in sections:
        find_and_fix_descriptors(doc, section)


def find_and_fix_descriptors(doc, container):
    children = container.childNodes
    for child in children:
        if child.nodeType == ELEMENT:
            tagName = child.tagName
            if tagName in DESCRIPTOR_ELEMENTS:
                rewrite_descriptor(doc, child)
            elif tagName == "subsection":
                find_and_fix_descriptors(doc, child)


def rewrite_descriptor(doc, descriptor):
    #
    # Do these things:
    #   1. Add an "index='no'" attribute to the element if the tagName
    #      ends in 'ni', removing the 'ni' from the name.
    #   2. Create a <signature> from the name attribute
    #   2a.Create an <args> if it appears to be available.
    #   3. Create additional <signature>s from <*line{,ni}> elements,
    #      if found.
    #   4. If a <versionadded> is found, move it to an attribute on the
    #      descriptor.
    #   5. Move remaining child nodes to a <description> element.
    #   6. Put it back together.
    #
    # 1.
    descname = descriptor.tagName
    index = descriptor.getAttribute("name") != "no"
    desctype = descname[:-4] # remove 'desc'
    linename = desctype + "line"
    if not index:
        linename = linename + "ni"
    # 2.
    signature = doc.createElement("signature")
    name = doc.createElement("name")
    signature.appendChild(doc.createTextNode("\n    "))
    signature.appendChild(name)
    name.appendChild(doc.createTextNode(descriptor.getAttribute("name")))
    descriptor.removeAttribute("name")
    # 2a.
    if descriptor.hasAttribute("var"):
        if descname != "opcodedesc":
            raise RuntimeError, \
                  "got 'var' attribute on descriptor other than opcodedesc"
        variable = descriptor.getAttribute("var")
        if variable:
            args = doc.createElement("args")
            args.appendChild(doc.createTextNode(variable))
            signature.appendChild(doc.createTextNode("\n    "))
            signature.appendChild(args)
        descriptor.removeAttribute("var")
    newchildren = [signature]
    children = descriptor.childNodes
    pos = skip_leading_nodes(children)
    if pos < len(children):
        child = children[pos]
        if child.nodeName == "args":
            # move <args> to <signature>, or remove if empty:
            child.parentNode.removeChild(child)
            if len(child.childNodes):
                signature.appendChild(doc.createTextNode("\n    "))
                signature.appendChild(child)
    signature.appendChild(doc.createTextNode("\n  "))
    # 3, 4.
    pos = skip_leading_nodes(children, pos)
    while pos < len(children) \
          and children[pos].nodeName in (linename, "versionadded"):
        if children[pos].tagName == linename:
            # this is really a supplemental signature, create <signature>
            oldchild = children[pos].cloneNode(1)
            try:
                sig = methodline_to_signature(doc, children[pos])
            except KeyError:
                print oldchild.toxml()
                raise
            newchildren.append(sig)
        else:
            # <versionadded added=...>
            descriptor.setAttribute(
                "added", children[pos].getAttribute("version"))
        pos = skip_leading_nodes(children, pos + 1)
    # 5.
    description = doc.createElement("description")
    description.appendChild(doc.createTextNode("\n"))
    newchildren.append(description)
    move_children(descriptor, description, pos)
    last = description.childNodes[-1]
    if last.nodeType == TEXT:
        last.data = last.data.rstrip() + "\n  "
    # 6.
    # should have nothing but whitespace and signature lines in <descriptor>;
    # discard them
    while descriptor.childNodes:
        descriptor.removeChild(descriptor.childNodes[0])
    for node in newchildren:
        descriptor.appendChild(doc.createTextNode("\n  "))
        descriptor.appendChild(node)
    descriptor.appendChild(doc.createTextNode("\n"))


def methodline_to_signature(doc, methodline):
    signature = doc.createElement("signature")
    signature.appendChild(doc.createTextNode("\n    "))
    name = doc.createElement("name")
    name.appendChild(doc.createTextNode(methodline.getAttribute("name")))
    methodline.removeAttribute("name")
    signature.appendChild(name)
    if len(methodline.childNodes):
        args = doc.createElement("args")
        signature.appendChild(doc.createTextNode("\n    "))
        signature.appendChild(args)
        move_children(methodline, args)
    signature.appendChild(doc.createTextNode("\n  "))
    return signature


def move_children(origin, dest, start=0):
    children = origin.childNodes
    while start < len(children):
        node = children[start]
        origin.removeChild(node)
        dest.appendChild(node)


def handle_appendix(doc, fragment):
    # must be called after simplfy() if document is multi-rooted to begin with
    docelem = get_documentElement(fragment)
    toplevel = docelem.tagName == "manual" and "chapter" or "section"
    appendices = 0
    nodes = []
    for node in docelem.childNodes:
        if appendices:
            nodes.append(node)
        elif node.nodeType == ELEMENT:
            appnodes = node.getElementsByTagName("appendix")
            if appnodes:
                appendices = 1
                parent = appnodes[0].parentNode
                parent.removeChild(appnodes[0])
                parent.normalize()
    if nodes:
        map(docelem.removeChild, nodes)
        docelem.appendChild(doc.createTextNode("\n\n\n"))
        back = doc.createElement("back-matter")
        docelem.appendChild(back)
        back.appendChild(doc.createTextNode("\n"))
        while nodes and nodes[0].nodeType == TEXT \
              and not nodes[0].data.strip():
            del nodes[0]
        map(back.appendChild, nodes)
        docelem.appendChild(doc.createTextNode("\n"))


def handle_labels(doc, fragment):
    for label in find_all_elements(fragment, "label"):
        id = label.getAttribute("id")
        if not id:
            continue
        parent = label.parentNode
        parentTagName = parent.tagName
        if parentTagName == "title":
            parent.parentNode.setAttribute("id", id)
        else:
            parent.setAttribute("id", id)
        # now, remove <label id="..."/> from parent:
        parent.removeChild(label)
        if parentTagName == "title":
            parent.normalize()
            children = parent.childNodes
            if children[-1].nodeType == TEXT:
                children[-1].data = children[-1].data.rstrip()


def fixup_trailing_whitespace(doc, fragment, wsmap):
    queue = [fragment]
    fixups = []
    while queue:
        node = queue[0]
        del queue[0]
        if wsmap.has_key(node.nodeName):
            fixups.append(node)
        for child in node.childNodes:
            if child.nodeType == ELEMENT:
                queue.append(child)

    # reverse the list to process from the inside out
    fixups.reverse()
    for node in fixups:
        node.parentNode.normalize()
        lastchild = node.lastChild
        before, after = wsmap[node.tagName]
        if lastchild.nodeType == TEXT:
            data = lastchild.data.rstrip() + before
            lastchild.data = data
        norm = 0
        if wsmap[node.tagName]:
            nextnode = node.nextSibling
            if nextnode and nextnode.nodeType == TEXT:
                nextnode.data = after + nextnode.data.lstrip()
            else:
                wsnode = doc.createTextNode(after)
                node.parentNode.insertBefore(wsnode, nextnode)
        # hack to get the title in place:
        if node.tagName == "title" \
           and node.parentNode.firstChild.nodeType == ELEMENT:
            node.parentNode.insertBefore(doc.createTextNode("\n  "),
                                         node.parentNode.firstChild)
            node.parentNode.normalize()


def normalize(doc):
    for node in doc.childNodes:
        if node.nodeType == ELEMENT:
            node.normalize()


def cleanup_trailing_parens(doc, element_names):
    d = {}
    for gi in element_names:
        d[gi] = gi
    rewrite_element = d.has_key
    queue = [node for node in doc.childNodes if node.nodeType == ELEMENT]
    while queue:
        node = queue[0]
        del queue[0]
        if rewrite_element(node.tagName):
            lastchild = node.lastChild
            if lastchild and lastchild.nodeType == TEXT:
                data = lastchild.data
                if data.endswith("()"):
                    lastchild.data = data[:-2]
        else:
            for child in node.childNodes:
                if child.nodeType == ELEMENT:
                    queue.append(child)


def contents_match(left, right):
    left_children = left.childNodes
    right_children = right.childNodes
    if len(left_children) != len(right_children):
        return 0
    for l, r in map(None, left_children, right_children):
        nodeType = l.nodeType
        if nodeType != r.nodeType:
            return 0
        if nodeType == ELEMENT:
            if l.tagName != r.tagName:
                return 0
            # should check attributes, but that's not a problem here
            if not contents_match(l, r):
                return 0
        elif nodeType == TEXT:
            if l.data != r.data:
                return 0
        else:
            # not quite right, but good enough
            return 0
    return 1


def create_module_info(doc, section):
    # Heavy.
    node = extract_first_element(section, "modulesynopsis")
    if node is None:
        return
    set_tagName(node, "synopsis")
    lastchild = node.childNodes[-1]
    if lastchild.nodeType == TEXT \
       and lastchild.data[-1:] == ".":
        lastchild.data = lastchild.data[:-1]
    modauthor = extract_first_element(section, "moduleauthor")
    if modauthor:
        set_tagName(modauthor, "author")
        modauthor.appendChild(doc.createTextNode(
            modauthor.getAttribute("name")))
        modauthor.removeAttribute("name")
    platform = extract_first_element(section, "platform")
    if section.tagName == "section":
        modinfo_pos = 2
        modinfo = doc.createElement("moduleinfo")
        moddecl = extract_first_element(section, "declaremodule")
        name = None
        if moddecl:
            modinfo.appendChild(doc.createTextNode("\n    "))
            name = moddecl.attributes["name"].value
            namenode = doc.createElement("name")
            namenode.appendChild(doc.createTextNode(name))
            modinfo.appendChild(namenode)
            type = moddecl.attributes.get("type")
            if type:
                type = type.value
                modinfo.appendChild(doc.createTextNode("\n    "))
                typenode = doc.createElement("type")
                typenode.appendChild(doc.createTextNode(type))
                modinfo.appendChild(typenode)
        versionadded = extract_first_element(section, "versionadded")
        if versionadded:
            modinfo.setAttribute("added", versionadded.getAttribute("version"))
        title = get_first_element(section, "title")
        if title:
            children = title.childNodes
            if len(children) >= 2 \
               and children[0].nodeName == "module" \
               and children[0].childNodes[0].data == name:
                # this is it; morph the <title> into <short-synopsis>
                first_data = children[1]
                if first_data.data[:4] == " ---":
                    first_data.data = first_data.data[4:].lstrip()
                set_tagName(title, "short-synopsis")
                if children[-1].nodeType == TEXT \
                   and children[-1].data[-1:] == ".":
                    children[-1].data = children[-1].data[:-1]
                section.removeChild(title)
                section.removeChild(section.childNodes[0])
                title.removeChild(children[0])
                modinfo_pos = 0
            else:
                ewrite("module name in title doesn't match"
                       " <declaremodule/>; no <short-synopsis/>\n")
        else:
            ewrite("Unexpected condition: <section/> without <title/>\n")
        modinfo.appendChild(doc.createTextNode("\n    "))
        modinfo.appendChild(node)
        if title and not contents_match(title, node):
            # The short synopsis is actually different,
            # and needs to be stored:
            modinfo.appendChild(doc.createTextNode("\n    "))
            modinfo.appendChild(title)
        if modauthor:
            modinfo.appendChild(doc.createTextNode("\n    "))
            modinfo.appendChild(modauthor)
        if platform:
            modinfo.appendChild(doc.createTextNode("\n    "))
            modinfo.appendChild(platform)
        modinfo.appendChild(doc.createTextNode("\n  "))
        section.insertBefore(modinfo, section.childNodes[modinfo_pos])
        section.insertBefore(doc.createTextNode("\n  "), modinfo)
        #
        # The rest of this removes extra newlines from where we cut out
        # a lot of elements.  A lot of code for minimal value, but keeps
        # keeps the generated *ML from being too funny looking.
        #
        section.normalize()
        children = section.childNodes
        for i in range(len(children)):
            node = children[i]
            if node.nodeName == "moduleinfo":
                nextnode = children[i+1]
                if nextnode.nodeType == TEXT:
                    data = nextnode.data
                    s = data.lstrip()
                    if len(s) < (len(data) - 4):
                        nextnode.data = "\n\n\n" + s


def cleanup_synopses(doc, fragment):
    for node in find_all_elements(fragment, "section"):
        create_module_info(doc, node)


def fixup_table_structures(doc, fragment):
    for table in find_all_elements(fragment, "table"):
        fixup_table(doc, table)


def fixup_table(doc, table):
    # create the table head
    thead = doc.createElement("thead")
    row = doc.createElement("row")
    move_elements_by_name(doc, table, row, "entry")
    thead.appendChild(doc.createTextNode("\n    "))
    thead.appendChild(row)
    thead.appendChild(doc.createTextNode("\n    "))
    # create the table body
    tbody = doc.createElement("tbody")
    prev_row = None
    last_was_hline = 0
    children = table.childNodes
    for child in children:
        if child.nodeType == ELEMENT:
            tagName = child.tagName
            if tagName == "hline" and prev_row is not None:
                prev_row.setAttribute("rowsep", "1")
            elif tagName == "row":
                prev_row = child
    # save the rows:
    tbody.appendChild(doc.createTextNode("\n    "))
    move_elements_by_name(doc, table, tbody, "row", sep="\n    ")
    # and toss the rest:
    while children:
        child = children[0]
        nodeType = child.nodeType
        if nodeType == TEXT:
            if child.data.strip():
                raise ConversionError("unexpected free data in <%s>: %r"
                                      % (table.tagName, child.data))
            table.removeChild(child)
            continue
        if nodeType == ELEMENT:
            if child.tagName != "hline":
                raise ConversionError(
                    "unexpected <%s> in table" % child.tagName)
            table.removeChild(child)
            continue
        raise ConversionError(
            "unexpected %s node in table" % child.__class__.__name__)
    # nothing left in the <table>; add the <thead> and <tbody>
    tgroup = doc.createElement("tgroup")
    tgroup.appendChild(doc.createTextNode("\n  "))
    tgroup.appendChild(thead)
    tgroup.appendChild(doc.createTextNode("\n  "))
    tgroup.appendChild(tbody)
    tgroup.appendChild(doc.createTextNode("\n  "))
    table.appendChild(tgroup)
    # now make the <entry>s look nice:
    for row in table.getElementsByTagName("row"):
        fixup_row(doc, row)


def fixup_row(doc, row):
    entries = []
    map(entries.append, row.childNodes[1:])
    for entry in entries:
        row.insertBefore(doc.createTextNode("\n         "), entry)
#    row.appendChild(doc.createTextNode("\n      "))


def move_elements_by_name(doc, source, dest, name, sep=None):
    nodes = []
    for child in source.childNodes:
        if child.nodeName == name:
            nodes.append(child)
    for node in nodes:
        source.removeChild(node)
        dest.appendChild(node)
        if sep:
            dest.appendChild(doc.createTextNode(sep))


RECURSE_INTO_PARA_CONTAINERS = (
    "chapter", "abstract", "enumerate",
    "section", "subsection", "subsubsection",
    "paragraph", "subparagraph", "back-matter",
    "howto", "manual",
    "item", "itemize", "fulllineitems", "enumeration", "descriptionlist",
    "definitionlist", "definition",
    )

PARA_LEVEL_ELEMENTS = (
    "moduleinfo", "title", "verbatim", "enumerate", "item",
    "interpreter-session", "back-matter", "interactive-session",
    "opcodedesc", "classdesc", "datadesc",
    "cfuncdesc", "ctypedesc", "cvardesc",
    "funcdesc", "methoddesc", "excdesc", "memberdesc", "membderdescni",
    "funcdescni", "methoddescni", "excdescni",
    "tableii", "tableiii", "tableiv", "localmoduletable",
    "sectionauthor", "seealso", "itemize",
    # include <para>, so we can just do it again to get subsequent paras:
    PARA_ELEMENT,
    )

PARA_LEVEL_PRECEEDERS = (
    "setindexsubitem", "author",
    "stindex", "obindex", "COMMENT", "label", "xi:include", "title",
    "versionadded", "versionchanged", "declaremodule", "modulesynopsis",
    "moduleauthor", "indexterm", "leader",
    )


def fixup_paras(doc, fragment):
    for child in fragment.childNodes:
        if child.nodeName in RECURSE_INTO_PARA_CONTAINERS:
            fixup_paras_helper(doc, child)
    descriptions = find_all_elements(fragment, "description")
    for description in descriptions:
        fixup_paras_helper(doc, description)


def fixup_paras_helper(doc, container, depth=0):
    # document is already normalized
    children = container.childNodes
    start = skip_leading_nodes(children)
    while len(children) > start:
        if children[start].nodeName in RECURSE_INTO_PARA_CONTAINERS:
            # Something to recurse into:
            fixup_paras_helper(doc, children[start])
        else:
            # Paragraph material:
            build_para(doc, container, start, len(children))
            if DEBUG_PARA_FIXER and depth == 10:
                sys.exit(1)
        start = skip_leading_nodes(children, start + 1)


def build_para(doc, parent, start, i):
    children = parent.childNodes
    after = start + 1
    have_last = 0
    BREAK_ELEMENTS = PARA_LEVEL_ELEMENTS + RECURSE_INTO_PARA_CONTAINERS
    # Collect all children until \n\n+ is found in a text node or a
    # member of BREAK_ELEMENTS is found.
    for j in range(start, i):
        after = j + 1
        child = children[j]
        nodeType = child.nodeType
        if nodeType == ELEMENT:
            if child.tagName in BREAK_ELEMENTS:
                after = j
                break
        elif nodeType == TEXT:
            pos = child.data.find("\n\n")
            if pos == 0:
                after = j
                break
            if pos >= 1:
                child.splitText(pos)
                break
    else:
        have_last = 1
    if (start + 1) > after:
        raise ConversionError(
            "build_para() could not identify content to turn into a paragraph")
    if children[after - 1].nodeType == TEXT:
        # we may need to split off trailing white space:
        child = children[after - 1]
        data = child.data
        if data.rstrip() != data:
            have_last = 0
            child.splitText(len(data.rstrip()))
    para = doc.createElement(PARA_ELEMENT)
    prev = None
    indexes = range(start, after)
    indexes.reverse()
    for j in indexes:
        node = parent.childNodes[j]
        parent.removeChild(node)
        para.insertBefore(node, prev)
        prev = node
    if have_last:
        parent.appendChild(para)
        parent.appendChild(doc.createTextNode("\n\n"))
        return len(parent.childNodes)
    else:
        nextnode = parent.childNodes[start]
        if nextnode.nodeType == TEXT:
            if nextnode.data and nextnode.data[0] != "\n":
                nextnode.data = "\n" + nextnode.data
        else:
            newnode = doc.createTextNode("\n")
            parent.insertBefore(newnode, nextnode)
            nextnode = newnode
            start = start + 1
        parent.insertBefore(para, nextnode)
        return start + 1


def skip_leading_nodes(children, start=0):
    """Return index into children of a node at which paragraph building should
    begin or a recursive call to fixup_paras_helper() should be made (for
    subsections, etc.).

    When the return value >= len(children), we've built all the paras we can
    from this list of children.
    """
    i = len(children)
    while i > start:
        # skip over leading comments and whitespace:
        child = children[start]
        nodeType = child.nodeType
        if nodeType == TEXT:
            data = child.data
            shortened = data.lstrip()
            if shortened:
                if data != shortened:
                    # break into two nodes: whitespace and non-whitespace
                    child.splitText(len(data) - len(shortened))
                    return start + 1
                return start
            # all whitespace, just skip
        elif nodeType == ELEMENT:
            tagName = child.tagName
            if tagName in RECURSE_INTO_PARA_CONTAINERS:
                return start
            if tagName not in PARA_LEVEL_ELEMENTS + PARA_LEVEL_PRECEEDERS:
                return start
        start = start + 1
    return start


def fixup_rfc_references(doc, fragment):
    for rfcnode in find_all_elements_from_set(fragment, ("pep", "rfc")):
        rfcnode.appendChild(doc.createTextNode(
            rfcnode.tagName.upper() + " " + rfcnode.getAttribute("num")))


def fixup_signatures(doc, fragment):
    for child in fragment.childNodes:
        if child.nodeType == ELEMENT:
            args = child.getElementsByTagName("args")
            for arg in args:
                rewrite_args(doc, arg)
            args = child.getElementsByTagName("constructor-args")
            for arg in args:
                rewrite_args(doc, arg)

def rewrite_args(doc, arglist):
    fixup_args(doc, arglist)
    arglist.normalize()
    if arglist.childNodes.length == 1 and arglist.firstChild.nodeType == TEXT:
        node = arglist.firstChild
        node.data = ' '.join(node.data.split())

def fixup_args(doc, arglist):
    for child in arglist.childNodes:
        if child.nodeName == "optional":
            # found it; fix and return
            arglist.insertBefore(doc.createTextNode("["), child)
            optkids = child.childNodes
            while optkids:
                arglist.insertBefore(child.firstChild, child)
            arglist.insertBefore(doc.createTextNode("]"), child)
            arglist.removeChild(child)
            return fixup_args(doc, arglist)


def fixup_sectionauthors(doc, fragment):
    for sectauth in find_all_elements(fragment, "sectionauthor"):
        section = sectauth.parentNode
        section.removeChild(sectauth)
        set_tagName(sectauth, "author")
        sectauth.appendChild(doc.createTextNode(
            sectauth.getAttribute("name")))
        sectauth.removeAttribute("name")
        after = section.childNodes[2]
        title = section.childNodes[1]
        if title.nodeName != "title":
            after = section.childNodes[0]
        section.insertBefore(doc.createTextNode("\n  "), after)
        section.insertBefore(sectauth, after)


def fixup_verbatims(doc):
    for verbatim in find_all_elements(doc, "verbatim"):
        child = verbatim.childNodes[0]
        if child.nodeType == TEXT \
           and child.data.lstrip().startswith(">>>"):
            set_tagName(verbatim, "interactive-session")


def add_node_ids(fragment, counter=0):
    fragment.node_id = counter
    for node in fragment.childNodes:
        counter = counter + 1
        if node.nodeType == ELEMENT:
            counter = add_node_ids(node, counter)
        else:
            node.node_id = counter
    return counter + 1


def fixup_ulink(doc, fragment):
    for ulink in find_all_elements(fragment, "ulink"):
        children = ulink.childNodes
        assert len(children) == 2
        text = children[0]
        href = children[1]
        href.normalize()
        assert len(href.childNodes) == 1
        assert href.childNodes[0].nodeType == TEXT
        url = href.childNodes[0].data
        ulink.setAttribute("href", url)
        ulink.removeChild(href)
        content = text.childNodes
        while len(content):
            ulink.appendChild(content[0])
        ulink.removeChild(text)


REFMODINDEX_ELEMENTS = ('refmodindex', 'refbimodindex',
                        'refexmodindex', 'refstmodindex')

def fixup_refmodindexes(fragment):
    # Locate <ref*modindex>...</> co-located with <module>...</>, and
    # remove the <ref*modindex>, replacing it with index=index on the
    # <module> element.
    nodes = find_all_elements_from_set(fragment, REFMODINDEX_ELEMENTS)
    d = {}
    for node in nodes:
        parent = node.parentNode
        d[parent.node_id] = parent
    del nodes
    map(fixup_refmodindexes_chunk, d.values())


def fixup_refmodindexes_chunk(container):
    # node is probably a <para>; let's see how often it isn't:
    if container.tagName != PARA_ELEMENT:
        bwrite("--- fixup_refmodindexes_chunk(%s)\n" % container)
    module_entries = find_all_elements(container, "module")
    if not module_entries:
        return
    index_entries = find_all_elements_from_set(container, REFMODINDEX_ELEMENTS)
    removes = []
    for entry in index_entries:
        children = entry.childNodes
        if len(children) != 0:
            bwrite("--- unexpected number of children for %s node:\n"
                   % entry.tagName)
            ewrite(entry.toxml() + "\n")
            continue
        found = 0
        module_name = entry.getAttribute("module")
        for node in module_entries:
            if len(node.childNodes) != 1:
                continue
            this_name = node.childNodes[0].data
            if this_name == module_name:
                found = 1
                node.setAttribute("index", "yes")
        if found:
            removes.append(entry)
    for node in removes:
        container.removeChild(node)


def fixup_bifuncindexes(fragment):
    nodes = find_all_elements(fragment, 'bifuncindex')
    d = {}
    # make sure that each parent is only processed once:
    for node in nodes:
        parent = node.parentNode
        d[parent.node_id] = parent
    del nodes
    map(fixup_bifuncindexes_chunk, d.values())


def fixup_bifuncindexes_chunk(container):
    removes = []
    entries = find_all_child_elements(container, "bifuncindex")
    function_entries = find_all_child_elements(container, "function")
    for entry in entries:
        function_name = entry.getAttribute("name")
        found = 0
        for func_entry in function_entries:
            t2 = func_entry.childNodes[0].data
            if t2[-2:] != "()":
                continue
            t2 = t2[:-2]
            if t2 == function_name:
                func_entry.setAttribute("index", "yes")
                func_entry.setAttribute("module", "__builtin__")
                if not found:
                    found = 1
                    removes.append(entry)
    for entry in removes:
        container.removeChild(entry)


def join_adjacent_elements(container, gi):
    queue = [container]
    while queue:
        parent = queue.pop()
        i = 0
        children = parent.childNodes
        nchildren = len(children)
        while i < (nchildren - 1):
            child = children[i]
            if child.nodeName == gi:
                if children[i+1].nodeName == gi:
                    ewrite("--- merging two <%s/> elements\n" % gi)
                    child = children[i]
                    nextchild = children[i+1]
                    nextchildren = nextchild.childNodes
                    while len(nextchildren):
                        node = nextchildren[0]
                        nextchild.removeChild(node)
                        child.appendChild(node)
                    parent.removeChild(nextchild)
                    continue
            if child.nodeType == ELEMENT:
                queue.append(child)
            i = i + 1


_token_rx = re.compile(r"[a-zA-Z][a-zA-Z0-9.-]*$")

def write_esis(doc, ofp, knownempty):
    for node in doc.childNodes:
        nodeType = node.nodeType
        if nodeType == ELEMENT:
            gi = node.tagName
            if knownempty(gi):
                if node.hasChildNodes():
                    raise ValueError, \
                          "declared-empty node <%s> has children" % gi
                ofp.write("e\n")
            for k, value in node.attributes.items():
                if _token_rx.match(value):
                    dtype = "TOKEN"
                else:
                    dtype = "CDATA"
                ofp.write("A%s %s %s\n" % (k, dtype, esistools.encode(value)))
            ofp.write("(%s\n" % gi)
            write_esis(node, ofp, knownempty)
            ofp.write(")%s\n" % gi)
        elif nodeType == TEXT:
            ofp.write("-%s\n" % esistools.encode(node.data))
        elif nodeType == ENTITY_REFERENCE:
            ofp.write("&%s\n" % node.nodeName)
        else:
            raise RuntimeError, "unsupported node type: %s" % nodeType


def convert(ifp, ofp):
    events = esistools.parse(ifp)
    toktype, doc = events.getEvent()
    fragment = doc.createDocumentFragment()
    events.expandNode(fragment)

    normalize(fragment)
    simplify(doc, fragment)
    handle_labels(doc, fragment)
    handle_appendix(doc, fragment)
    fixup_trailing_whitespace(doc, fragment, {
        # element -> (before-end-tag, after-end-tag)
        "abstract": ("\n", "\n"),
        "title": ("", "\n"),
        "chapter": ("\n", "\n\n\n"),
        "section": ("\n", "\n\n\n"),
        "subsection": ("\n", "\n\n"),
        "subsubsection": ("\n", "\n\n"),
        "paragraph": ("\n", "\n\n"),
        "subparagraph": ("\n", "\n\n"),
        "description": ("\n", "\n\n"),
        "enumeration": ("\n", "\n\n"),
        "item": ("\n", "\n\n"),
        })
    cleanup_root_text(doc)
    cleanup_trailing_parens(fragment, ["function", "method", "cfunction"])
    cleanup_synopses(doc, fragment)
    fixup_descriptors(doc, fragment)
    fixup_verbatims(fragment)
    normalize(fragment)
    fixup_paras(doc, fragment)
    fixup_sectionauthors(doc, fragment)
    fixup_table_structures(doc, fragment)
    fixup_rfc_references(doc, fragment)
    fixup_signatures(doc, fragment)
    fixup_ulink(doc, fragment)
    add_node_ids(fragment)
    fixup_refmodindexes(fragment)
    fixup_bifuncindexes(fragment)
    # Take care of ugly hacks in the LaTeX markup to avoid LaTeX and
    # LaTeX2HTML screwing with GNU-style long options (the '--' problem).
    join_adjacent_elements(fragment, "option")
    # Attempt to avoid trailing blank lines:
    fragment.normalize()
    if fragment.lastChild.data[-1:] == "\n":
        fragment.lastChild.data = fragment.lastChild.data.rstrip() + "\n"
    #
    d = {}
    for gi in events.parser.get_empties():
        d[gi] = gi
    for key in ("author", "pep", "rfc"):
        if d.has_key(key):
            del d[key]
    knownempty = d.has_key
    #
    try:
        write_esis(fragment, ofp, knownempty)
    except IOError, (err, msg):
        # Ignore EPIPE; it just means that whoever we're writing to stopped
        # reading.  The rest of the output would be ignored.  All other errors
        # should still be reported,
        if err != errno.EPIPE:
            raise


def main():
    if len(sys.argv) == 1:
        ifp = sys.stdin
        ofp = sys.stdout
    elif len(sys.argv) == 2:
        ifp = open(sys.argv[1])
        ofp = sys.stdout
    elif len(sys.argv) == 3:
        ifp = open(sys.argv[1])
        import StringIO
        ofp = StringIO.StringIO()
    else:
        usage()
        sys.exit(2)
    convert(ifp, ofp)
    if len(sys.argv) == 3:
        fp = open(sys.argv[2], "w")
        fp.write(ofp.getvalue())
        fp.close()
        ofp.close()


if __name__ == "__main__":
    main()



from __future__ import absolute_import

import os
import re
import subprocess
from datetime import datetime, timedelta

import GangaCore.Utility.logging
from GangaCore.Core.exceptions import InvalidCredentialError, GangaValueError
from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Schema import SimpleItem

from GangaCore.GPIDev.Adapters.ICredentialInfo import cache, retry_command
from GangaCore.GPIDev.Adapters.ICredentialRequirement import ICredentialRequirement
from GangaCore.Core.exceptions import CredentialRenewalError
from GangaCore.GPIDev.Credentials.VomsProxy import VomsProxyInfo
from GangaCore.Utility.Shell import Shell

from GangaDirac.Lib.Utilities.DiracUtilities import getDiracEnv

logger = GangaCore.Utility.logging.getLogger()

class DiracProxyInfo(VomsProxyInfo):
    """
    A wrapper around a DIRAC proxy file
    """

    def __init__(self, requirements, check_file=False, create=False):
        """
        Args:
            requirements (ICredentialRequirement): An object specifying the requirements
            check_file (bool): Raise an exception if the file does not exist
            create (bool): Create the credential file
        """
        self._shell = Shell()

        super(DiracProxyInfo, self).__init__(requirements, check_file, create)

    @retry_command
    def create(self):
        """
        Creates the grid proxy.

        Raises:
            CredentialRenewalError: If the renewal process returns a non-zero value
        """
        group_command = ''
        logger.debug('require ' + self.initial_requirements.group)
        if self.initial_requirements.group:
            group_command = '--group %s --VOMS' % self.initial_requirements.group
        validTime_command = ''
        if self.initial_requirements.validTime:
            r = re.compile('\d{2}:\d{2}$')
            if r.match(self.initial_requirements.validTime):
                validTime_command = '--valid %s' % self.initial_requirements.validTime
            else:
                logger.error('Supplied time for validation not of correct format "HH:MM". Failed to create DIRAC proxy')
                raise CredentialRenewalError('Supplied time for validation not of correct format "HH:MM". Failed to create DIRAC proxy')
        command = getConfig('DIRAC')['proxyInitCmd'] + ' --strict --out "%s" %s %s' % (self.location, group_command, validTime_command)
        logger.debug(command)
        self.shell.env['X509_USER_PROXY'] = self.location
        try:
            self.shell.check_call(command)
        except subprocess.CalledProcessError:
            raise CredentialRenewalError('Failed to create DIRAC proxy')
        else:
            logger.debug('Grid proxy {path} created. Valid for {time}'.format(path=self.location, time=self.time_left()))

    @property
    def shell(self):
        """
        Construct and store a shell which has the appropriate DIRAC env saved in it
        """
        if self._shell is None:
            self._shell = Shell()
            self._shell.env.update(getDiracEnv(self.initial_requirements.dirac_env))
        return self._shell

    def destroy(self):
        """
        This removes a given dirac proxy from disk
        """
        if os.path.isfile(self.location):
            os.remove(self.location)

    @cache
    def info(self):
        """
        Return the info on a given proxy using the dirac tools
        """
        self.shell.env['X509_USER_PROXY'] = self.location
        info_cmd = getConfig('DIRAC')['proxyInfoCmd'] + ' --file "%s"' % self.location
        logger.debug(info_cmd)
        status, output, message = self.shell.cmd1(info_cmd)
        return output

    def field(self, label):
        # type: (str) -> str
        """
        Returns the value for a given label out of self.info()
        self.info() returns data which looks like:
          key : value

        this method uses regex to return the value for a given key (label)
        Args:
            label (str): Label which is expected to describe some property of the dirac proxy
        """
        line = re.search(r'^{0}\s*: (.*)$'.format(label), self.info(), re.MULTILINE)
        if line is None:
            raise InvalidCredentialError()
        return line.group(1)

    @property
    @cache
    def identity(self):
        """
        Returns the identity associated with the dirac proxy
        """
        return self.field('identity')

    @property
    @cache
    def group(self):
        """
        Returns the group associated with the dirac proxy
        """
        return self.field('DIRAC group')

    @property
    @cache
    def username(self):
        """
        Returns the username associated with the dirac proxy
        """
        return self.field('username')

    @property
    def validTime(self):
        return self.initial_requirements.validTime

    @property
    def encodeDefaultProxyFileName(self):
        """
        Returns whether the proxy has any information encoded in the proxy filename
        """
        return self.initial_requirements.encodeDefaultProxyFileName

    @cache
    def expiry_time(self):
        """
        Returns the time in the future when the proxy will expire in seconds
        """
        time = self.field('timeleft')
        split_time = time.split(':')
        return datetime.now() + timedelta(hours=int(split_time[0]), minutes=int(split_time[1]), seconds=int(split_time[2]))

    def default_location(self):
        """
        Returns the default location of the proxy on disk including the encoded extension if it should be there
        """
        base_proxy_name = os.getenv('X509_USER_PROXY') or '/tmp/x509up_u' + str(os.getuid())
        encoded_ext = self.initial_requirements.encoded()
        if encoded_ext:
            return base_proxy_name + ':' + encoded_ext
        else:
            return base_proxy_name

    @property
    def location(self):
        """
        """
        return self.default_location()

class DiracProxy(ICredentialRequirement):
    """
    An object specifying the requirements of a DIRAC proxy file
    """
    _schema = ICredentialRequirement._schema.inherit_copy()
    _schema.datadict['group'] = SimpleItem(defvalue=None, typelist=[str, None], doc='Group for the proxy')
    _schema.datadict['encodeDefaultProxyFileName'] = \
        SimpleItem(defvalue=True, doc='Should the proxy be generated with the group encoded onto the end of the proxy filename')
    _schema.datadict['dirac_env'] = SimpleItem(defvalue=None, typelist=[str, None], doc='File which can be used to access a different DIRAC backend')
    _schema.datadict['validTime'] = SimpleItem(defvalue=None, typelist=[str, None], doc='Time for which proxy will be valid. Default if None is 24 hours. Must be of form "HH:MM"')
    _category = 'CredentialRequirement'

    info_class = DiracProxyInfo

    def __init__(self, **kwargs):
        """
        Constructs a dirac proxy requirement and assigns the default group if none is provided
        """
        super(DiracProxy, self).__init__(**kwargs)
        if self.group is None:
            raise GangaValueError('DIRAC Proxy `group` is not set. Set this in ~/.gangarc in `[defaults_DiracProxy]/group`')

    def encoded(self):
        """
        This returns the encoding used to store a unique DIRAC proxy for each group
        """
        my_config = getConfig('defaults_DiracProxy')
        default_group = my_config['group']
        if (my_config['encodeDefaultProxyFileName'] and self.group == default_group) or self.group != default_group:
            if self.dirac_env is not None:
                return ':'.join(requirement for requirement in [self.group] if requirement) + ':' + str(hash(self.dirac_env)) # filter out the empties
            else:
                return ':'.join(requirement for requirement in [self.group] if requirement)  # filter out the empties
        else:
            return ''

# A single global check for the DIRAC group setting. This will bail out early and safely during plugin loading.
if getConfig('defaults_DiracProxy')['group'] is None:
    raise GangaValueError('DIRAC Proxy `group` is not set. Set this in ~/.gangarc in `[defaults_DiracProxy]/group`')




from __future__ import absolute_import
import glob
import pickle
import os
import sys
import copy
from GangaCore.GPIDev.Persistency import load, stripped_export
from GangaCore.Core.GangaRepository import allRegistries, RepositoryError
from .GangaRepository import GangaRepository
from GangaCore.Utility.logging import getLogger
logger = getLogger()


class GangaRepositoryImmutableTransient(GangaRepository):

    def __init__(self, registry, filebase, file_ext='tpl', pickle_files=False, locking=True):
        """GangaRepository constructor. Initialization should be done in startup()"""
        super(GangaRepositoryImmutableTransient, self).__init__(registry)
        self.filebase = filebase
        self._next_id = 0
        self.file_ext = file_ext
        self.pickle_files = pickle_files
        self.registry = registry

    def startup(self):
        def _readonly(): return True

        # this is needed as the default registry that would be added to with Jebtemplates is the
        # templates registry and for Tasks would be the tasks registry
        # we put this back after loading. Note that the _auto__init for job.py
        # also calls for the prep registry so we have to be able to return this
        # one.
        def getRegistry(name):
            if name == 'prep':
                return allRegistries['prep']
            return self.registry
        old = getattr(sys.modules['GangaCore.Core.GangaRepository'], 'getRegistry')
        setattr(sys.modules['GangaCore.Core.GangaRepository'], 'getRegistry', getRegistry)

        # by setting the registry started now the auto_init from the jobTemplate class
        # call call getRegistry(self.default_registry)._add
        self.registry._hasStarted = True

        for f in glob.glob(os.path.join(self.filebase, '*.%s' % self.file_ext)):
            current_id = self._next_id
            try:
                if self.pickle_files:
                    obj = pickle.load(open(f, 'rb'))
                else:
                    from GangaCore.GPIDev.Base.Proxy import stripProxy
                    obj = stripProxy(load(f)[0])
            except:
                logger.error("Unable to load file '%s'" % f)
                setattr(
                    sys.modules['GangaCore.Core.GangaRepository'], 'getRegistry', old)
                raise
            else:
                obj.name = os.path.basename(f).rsplit('.', 1)[0]
                # if this not true then add already called from _auto__init
                # when loading the object. note default _auto__init is just
                # pass
                if self._next_id == current_id:
                    obj.id = self._next_id
                    obj._registry = self.registry
                    obj._registry_id = self._next_id
                    setattr(obj, '_readonly', _readonly)

                    self.objects[self._next_id] = obj
                    self._next_id += 1

        setattr(sys.modules['GangaCore.Core.GangaRepository'], 'getRegistry', old)

    def updateLocksNow(self):
        pass

    def update_index(self, id=None):
        pass

    def shutdown(self):
        pass

    def add(self, objs, force_ids=None):
        ids = []

        def _readonly(): return True
        for o in objs:
            obj = copy.deepcopy(o)
            fn = os.path.join(self.filebase, '%s.%s' %
                              (obj.name, self.file_ext))
            try:
                if self.pickle_files:
                    obj._registry = None

                    pickle.dump(obj, open(fn, 'wb'))

                else:
                    if not stripped_export(obj, fn):
                        raise RepositoryError(self, 'Failure in stripped_export method, returned False')
            except:
                logger.error("Unable to write to file '%s'" % fn)
                raise
            else:
                obj.id = self._next_id
                obj._registry = self.registry
                obj._registry_id = self._next_id
                setattr(obj, '_readonly', _readonly)

                self.objects[self._next_id] = obj
                ids.append(self._next_id)
                self._next_id += 1
        return ids

    def delete(self, ids):
        pass

    def load(self, ids):
        pass

    def flush(self, ids):
        pass

    def lock(self, ids):
        return True

    def unlock(self, ids):
        pass



import unittest
import traceback
import StringIO

from GangaCore.Utility.logging import getLogger
logger = getLogger(modulename=True)

class GangaGPITestCase(unittest.TestCase):
        """
        Base class for GPI test-cases
        """
        def __init__(self, methodName='runTest'):
                unittest.TestCase.__init__(self,methodName)


        def setUp(self):
                pass

        def tearDown(self):
                pass

        def runTest():
                pass

from pytf.lib import MultipassTest as MP 
class MultipassTest(MP):
        """
        Base class for multi-pass tests
        """
        def setUp(self):
                pass

        def tearDown(self):
                pass


class FailedTestsException(Exception):
    """The exception shows the failed tests"""
    def __init__(self, failedTests):
        self.failedTests = failedTests

    def __str__(self):
        msg = "\n**** FAIL ****\n\n"
        for name in self.failedTests:
            msg += "%s\nFAIL: %s, please check the error message below\n%s" % (('-' * 40), name[0], name[1])
        msg += '%s' % ('-' * 40)
        return "%s\n**** END ****" % msg

class GangaGPIPTestCase(unittest.TestCase):
    """
    The parent test case of the REAL GPIP tests.
    """
    def __init__(self, testCmd, descr, output_path, report_outputpath, parent_report_name, releaseTesting=False):
        unittest.TestCase.__init__(self,'run_pyunit_testcases')
        self.testCmd = testCmd
        self.description = descr
        self.enableReleaseTesting = releaseTesting
        self.output_path_template = output_path
        self.output_path = output_path % 'ALL'
        self.report_outputpath = report_outputpath
        self.parent_report_name = parent_report_name
        self._initializeTestCmd()

    def _initializeTestCmd(self):
        self.testCmd = '%s --output-path=%s --description=\'%s\'' % (self.testCmd, self.output_path_template, self.description.replace('ALL', '%s'))
        if self.enableReleaseTesting:
            self.testCmd = '%s %s' % (self.testCmd, '--release-test=True')
            self.testCmd = '%s --report-outputpath=%s' % (self.testCmd, self.report_outputpath)
            self.testCmd = '%s --parent-report-name=%s' % (self.testCmd, self.parent_report_name)

    def getDescription(self):
        return self.description

    def shortDescription(self):
        return self.description

    def isReleaseTesting(self):
        self.enableReleaseTesting = True

    def __str__(self):
        return self.description

    def run_pyunit_testcases(self):
        import os
        import sys
        from subprocess import PIPE,Popen,STDOUT
        logger.info('\n***** %s' % self.description)
        logger.info('\tLog file: %s' % self.output_path)

        output = open(self.output_path,'w')
        pytf_paths = str.split(os.getenv('PYTF_TOP_DIR'),':')
        pytf_runner = str()
        if type(pytf_paths) is str:
            pytf_runner = os.path.join( pytf_paths, 'cmd_run.sh' )
        else:
            for i in pytf_paths:
                if os.path.isfile( os.path.join( i, 'cmd_run.sh') ):
                    pytf_runner = os.path.join( i, 'cmd_run.sh')
                    break;
        script_runner = [pytf_runner,'{ ' + self.testCmd + '; }']
        process = Popen(script_runner, shell=False, bufsize=0,
            stdin=PIPE, stdout=output, stderr=STDOUT, close_fds=True)
        process.wait()
        output.close()

        
        text = None
        f = open(self.output_path)
        try: text = f.read()
        finally: f.close()

        start_index = text.find("**** FAIL ****")
        end_index = text.find("**** END ****")
        if start_index > -1:start_index = start_index + 15
        err = "%s" % text[start_index:end_index]
        if len(err) > 0:
            raise unittest.TestCase.failureException(err)

import time

class GPIPPreparationTestCase(unittest.TestCase):
    """
    The test case for running the test preparation.
    """
    def __init__(self, pytest_instance, description, method_name, timeout, outputpath):
        unittest.TestCase.__init__(self,'run_pytest_function')
        self.pytest_instance = pytest_instance
        self.method_name = method_name
        self.timeout = timeout
        self.outputpath = outputpath
        self.description = description #'%s_preparation' % description
        self.preparationError = None
        #self.description_org = description

    def run_pytest_function(self):
        logger.info('%s starts to run' % self.method_name)
        try:
            checkTest = getattr(self.pytest_instance, self.method_name)()
        except Exception as preparationError:
            sio = StringIO.StringIO()
            traceback.print_exc(file=sio)
            self.preparationError = sio.getvalue()
            sio.close()
            raise preparationError

        if checkTest:
            self.checkTest = checkTest

    def getCheckTest(self):
        try:
            return self.checkTest
        except AttributeError:
            raise Exception('Failed to get the instance of "CheckTest".(Should get this after running the testcase.)')

    def getDescription(self):
        return self.description

    def shortDescription(self):
        return self.description

    def __str__(self):
        return self.description


class GPIPCheckTestCase(unittest.TestCase):
    """
    The test case for running the result check.
    """
    def __init__(self, checkTest, description, preError=None, timeoutError=None):
        unittest.TestCase.__init__(self,'runCheckTest')
        self.checkTest = checkTest
        self.description = description
        self.preError = preError
        self.timeoutError = timeoutError
        self.runCheckError = None
        self.errorTraceback = None

    def runCheckTest(self):
        try:
            if self.timeoutError:
                raise self.timeoutError

            if self.preError:
                raise self.preError

            assert(self.checkTest is not None), 'No instance of checktest, this should happen while the preparation of test is failed.'
        except Exception as error:
            sio = StringIO.StringIO()
            traceback.print_exc(file=sio)
            self.errorTraceback = sio.getvalue()
            sio.close()
            raise error

        try:
            self.checkTest.checkTest()
        except Exception as runCheckError:
            self.runCheckError = runCheckError
            sio = StringIO.StringIO()
            traceback.print_exc(file=sio)
            self.errorTraceback = sio.getvalue()
            sio.close()
            raise runCheckError

    def getDescription(self):
        return self.description

    def shortDescription(self):
        return self.description

    def __str__(self):
        return self.description

from GangaTest.Framework.utils import is_job_completed
class ICheckTest(object):
    """
    The default ICheckTest interface implementation.
    """
    def __init__(self, j):
        self.j = j
    """Interface class for getting the status of result and checking result"""
    def checkTest(self):
        """This checkTest method will be invoked by the driver"""
        assert(self.j.status == 'completed')
    def isReadyForCheck(self):
        """This check method will be invoked by the driver"""
        return is_job_completed(self.j)

from pytf.testoob.running.convenience import apply_decorators
from pytf.testoob.extracting import full_extractor as _full_extractor

class SimpleRunnerControl(object):
    """
    The instance of this class will control one real GPIP test case, for running "prepare" -> "monitoring" -> "check result"
    """
    def __init__(self, testName, runner, preparationTestCase, fixture_decorators, timeout):
        self.testName = testName
        self.runner = runner
        self.preparationTestCase = preparationTestCase
        self.fixture_decorators = fixture_decorators
        self.timeout = timeout
        self.isTimeout = False
        self.error = None
        self.timeoutError = None
        self.finished = False
        self.preparationError = None
        self.runCheckError = None
        self.errorTraceback = None

    def runPreparationTestCase(self):
        self._printBegin('runPreparation')
        run_status =  self.runner.run(self.preparationTestCase)
        if self.preparationTestCase.get_fixture().preparationError:
            self.preparationError = self.preparationTestCase.get_fixture().preparationError
        try:
            self.checkTest = self.preparationTestCase.get_fixture().getCheckTest()
        except Exception as e:
            logger.warning("Failed to get the instance of check test, because of the fail of the preparation of test, won't do the check test.")
            self.checkTest = None 
            if run_status:
                logger.warning('The test preparation is done successfully, did you forget to return the instance of check test?')
            run_status = False # Mark the run status as fail, in order to set to finish by driver.py.
        self._printEnd('runPreparation')
        return run_status

    def isFinished(self):
        return self.finished

    def setFinished(self, finished=False):
        self.runner.done()
        self.finished = finished

    def isReadyForCheck(self):
         ready = False
         try:
             if self.checkTest:
                 ready = self.checkTest.isReadyForCheck()
             else:
                 ready = True
         except Exception as e:
             #Change to catch generic exception.
             sio = StringIO.StringIO()
             traceback.print_exc(file=sio)
             self.error = sio.getvalue()
             sio.close()
             #raise e

         #if ready is None:
         #    return False
         #else:
         return ready

    def getError(self):
         if self.erorr:
             return self.error
         else:
             return None

    def setTimeout(self, flag=False):
        self.isTimeout = flag

    def runCheckTest(self):
        run_status = False
        self._printBegin('runCheckTest')
        if self.isTimeout:
            self.timeoutError = unittest.TestCase.failureException("Test Timeout(%s sec). Consider increasing the value of '[TestingFramework]timeout' parameter in your test configuration file" % self.timeout)

        #if self.error is not None:

        checkTestCase = GPIPCheckTestCase(self.checkTest, "CheckTest", self.error, self.timeoutError)
        checkTestSuite = unittest.TestSuite([checkTestCase])

        test_extractor = apply_decorators(_full_extractor, [])

        for fixture in test_extractor(checkTestSuite):
            decorated_fixture = apply_decorators(fixture, self.fixture_decorators)
            run_status = self.runner.run(decorated_fixture)

        if checkTestCase.runCheckError:
            self.runCheckError = checkTestCase.runCheckError

        if checkTestCase.errorTraceback:
            self.errorTraceback = checkTestCase.errorTraceback

        self.runner.done()
        self._printEnd('runCheckTest')
        return run_status

    def _printBegin(self, method):
        logger.info('@@@@ [%s] BEGIN of %s' % (self.testName, method))

    def _printEnd(self, method):
        logger.info('@@@@ [%s] END of %s' % (self.testName, method))



#XML differencer - reads the XML files for two test versions
# and looks for differences between test case success failures
from GangaCore.Utility.Config import getConfig
import os, xml, fnmatch
from GangaCore.Utility.logging import getLogger
logger = getLogger()

newreportdir = None
newversion = None
newconfig = None
oldversion = None
oldconfig = None

def writeXMLDifferenceFile(newtests, oldtests, filename):
    global newreportdir, newversion, oldversion, newconfig, oldconfig
    #Initialise XMLWriter
    from pytf.testoob.compatibility.SimpleXMLWriter import XMLWriter
    #strip relevant part of filename
    ind = filename.find("__")
    filename = filename[:ind].strip()
    #write new difference file
    differencefile = os.path.join(newreportdir, filename+"__Diff_"+newversion+"-"+oldversion+"_"+newconfig+"_"+oldconfig+".xml")
    logger.info("writing:"+differencefile)
    #differencefile = "/home/alex/differencefile.xml"
    f = open(differencefile, 'w')
    w = XMLWriter(f)
    w.start('results')
    w.start('testsuites')
    #Now we have two dicts - now to match keys
    for test in newtests:
        testcase_new = newtests[test]
        try: testcase_old = oldtests[test]
        except KeyError: 
            w.start('testcase', {'name':testcase_new.attributes.items()[0][1]} )
            result = testcase_new.getElementsByTagName("result")[0].childNodes[0].data
            w.element('result', result )
            w.start(result)
            w.data("Brand New Test")
            w.end(result)
            w.end('testcase')
            continue 
        #work out what to do
        result_new = testcase_new.getElementsByTagName("result")[0].childNodes[0].data
        result_old = testcase_old.getElementsByTagName("result")[0].childNodes[0].data
        test_name = testcase_new.attributes.items()[0][1]
        test_time = testcase_new.attributes.items()[1][1]+" / "+testcase_old.attributes.items()[1][1]
        if ( result_new == 'failure' and result_old == 'failure' ):
            test_result = 'success'
            #write XML element
            w.start('testcase', { 'name':test_name , 'time':test_time } )
            w.element('result',test_result)
            w.start('failure', { 'message':'Both tests failed' , 'type': 'F->F'} )
            #w.start('success', { 'message':'Both tests failed' , 'type': 'F->F'} )
            #failureNode = testcase_new.getElementsByTagName("failure")
            #new_data = "Latest test: \n "+failureNode[0].childNodes[0].data+"\n"
            #failureNode = testcase_old.getElementsByTagName("failure")
            #old_data = "Previous test: \n "+failureNode[0].childNodes[0].data+"\n"                         
            #w.data(new_data+" "+"\n"+old_data)
            #w.end('success')
            w.end('failure')
            w.end('testcase')
        elif ( result_new == 'failure' and result_old == 'success' ): 
            test_result = 'failure'
            failureNode = testcase_new.getElementsByTagName("failure")
            new_data = "Changed from success to failure: \n "+failureNode[0].childNodes[0].data+"\n"+" - "+"oops"
            w.start('testcase', { 'name':test_name , 'time':test_time } )
            w.element('result',test_result)
            w.start('failure',{ 'message':'New test failed, old one passed' , 'type': 'S->F'} )
            w.data(new_data)
            w.end('failure')
            w.end('testcase')
        elif ( result_new == 'success' and result_old == 'failure' ):
            test_result = 'failure'
            old_data = "Changed from failure to success in new release"+" - "+"correct"
            w.start('testcase', { 'name':test_name , 'time':test_time } )
            w.element('result',test_result)
            w.start( 'success', { 'message':'Old test failed, new one passed' , 'type': 'F->S'} )
            #w.start('failure', { 'message':'Old test failed, new one passed' , 'type': 'F->S'} )
            w.data(old_data)
            #w.end('failure')
            w.end('success')
            w.end('testcase')
        elif ( result_new == 'success' and result_old == 'success' ):
            test_result = 'success'
            w.start('testcase', { 'name':test_name , 'time':test_time } )
            w.element('result',test_result)
            w.end('testcase')
        else:
            pass
        
    #close XMLWriter methods
    #w.flush()
    w.end('testsuites')
    w.end('results')
    f.close()
    
    
#======================================================================
def comparetestfiles(newfiledict, oldfiledict):
    #go through list of tests
    for filename in newfiledict:

        logger.info( "Looking at: %s" % filename )

        if filename not in oldfiledict:
            logger.info( "NOT Comparing: %s" % filename )
            continue
        logger.info( "Comparing: %s" % filename )

        newtests = {}
        oldtests = {}
        #go through new report
        newdoc = newfiledict[filename]
        testcases = newdoc.getElementsByTagName("testcase")
        for testcase in testcases:
           if testcase.nodeType == testcase.ELEMENT_NODE:
                for (name, value) in testcase.attributes.items():
                    if name == 'name':
                        #testcase_name=value.split()[0].split("/")
                        ind = value.find("[")
                        test = value[:ind].strip()
                #add to dictionary
                newtests[test] = testcase
        # Go through old report
        olddoc = oldfiledict[filename]
        testcases = olddoc.getElementsByTagName("testcase")
        for testcase in testcases:
           if testcase.nodeType == testcase.ELEMENT_NODE:
                for (name, value) in testcase.attributes.items():
                    if name == 'name':
                        #testcase_name=value.split()[0].split("/")
                        ind = value.find("[")
                        test = value[:ind].strip()
                #add to dictionary
                oldtests[test] = testcase
        #        
        #Call writeXMLfile and pass the two dicts
        writeXMLDifferenceFile(newtests, oldtests, filename)

#======================================================================
def decide_new(datalist):
    global newreportdir, newversion, oldversion, newconfig, oldconfig
    version_one = datalist[0]
    version_two = datalist[2]
    config_one = datalist[1]
    config_two = datalist[3]
    a = version_one.split(".")
    b = version_two.split(".")
    
    if ( a[0] == b[0] ):
        if ( a[1] == b[1] ):
            if ( a[2] == b[2] ):
                newversion = version_one
                oldversion = version_two
                newconfig = config_one
                oldconfig = config_two                        
            elif ( a[2] < b[2] ):
                newversion = version_two
                oldversion = version_one
                newconfig = config_two
                oldconfig = config_one
            elif ( a[2] > b[2] ):
                newversion = version_one
                oldversion = version_two
                newconfig = config_one
                oldconfig = config_two        
        elif ( a[1] < b[1] ):
            newversion = version_two
            oldversion = version_one
            newconfig = config_two
            oldconfig = config_one
        elif ( a[1] > b[1] ):
            newversion = version_one
            oldversion = version_two
            newconfig = config_one
            oldconfig = config_two        
    elif ( a[0] < b[0] ):
        newversion = version_two
        oldversion = version_one
        newconfig = config_two
        oldconfig = config_one
    elif ( a[0] > b[0] ):
        newversion = version_one
        oldversion = version_two
        newconfig = config_one
        oldconfig = config_two        

    return [newversion, newconfig, oldversion, oldconfig]


#======================================================================
def start(cmd_args=None):    
    #import global variables
    global newreportdir, newversion, oldversion, newconfig, oldconfig
    #
    config = getConfig('System')
    if cmd_args:
        logger.info(cmd_args)
    else:
        logger.error("no args passed")
        return

    logger.info( "Starting postprocessing" )

    #process args
    if ( len(cmd_args) == 1 ):
        
        search_dir = '/afs/cern.ch/sw/ganga/install/'

        def isPre(x):
            return (x.endswith('-pre') and (x.find('hotfix') == -1) and (x.find('beta') == -1))

        import os
        os.chdir(search_dir)
        dirs = filter(os.path.isdir, os.listdir(search_dir))
        dirs = filter(isPre, os.listdir(search_dir))
        dirs = [os.path.join(search_dir, d) for d in dirs] # add path to each dir
        dirs.sort(lambda x,y: cmp(os.path.getmtime(x),os.path.getmtime(y)))

        newversion = os.path.basename(dirs[-1])
        if len(dirs) > 1:
            oldversion = os.path.basename(dirs[-2])
        else:
            oldversion = newversion

        oldconfig = 'ALL'
        newconfig = 'ALL'
    elif ( len(cmd_args) == 2 ):
        # is of format: newversion, oldversion
        newversion = cmd_args[0]
        oldversion = cmd_args[1]        
        oldconfig = 'ALL'
        newconfig = 'ALL'
    elif ( len(cmd_args) == 3 ):
        #is of format: number, config, config
        oldversion = cmd_args[0]
        thisversion = config['GANGA_VERSION']
        a = []
        for i in range(len(thisversion)):
            if (thisversion[i] == '-'):
                a += thisversion[i+1]
        newversion = a[0]+'.'+a[1]+'.'+a[2] 
        oldconfig = cmd_args[1]
        newconfig = cmd_args[2]
    elif ( len(cmd_args) == 2 ):
        #is of format: number, number, config, config
        newversion = cmd_args[0]
        oldversion = cmd_args[1]        
        oldconfig = cmd_args[2]
        newconfig = cmd_args[3]
    
    datalist = decide_new([newversion, newconfig, oldversion, oldconfig])
    newversion = datalist[0]
    newconfig = datalist[1]
    oldversion = datalist[2]
    oldconfig = datalist[3]
    logger.info( "new version = "+newversion )
    logger.info( "oldversion = "+oldversion )
    
    #Get Newreportdir and oldreportdir
    topdir = os.sep.join(config['GANGA_PYTHONPATH'].split(os.sep)[:-2])
    for this_dir in os.listdir(topdir):
        if ( this_dir == newversion ):
            newreportdir = os.path.join(topdir,this_dir,'reports','latest')
        if ( this_dir == oldversion ):
            oldreportdir = os.path.join(topdir,this_dir,'reports','latest')
    try:
       logger.info("newreportdir = "+newreportdir)
    except:
       logger.warning("new version: "+str(newversion)+" not found in:"+str(newreportdir))
       return
    try:
        logger.info("oldreportdir = "+oldreportdir)
    except:
        logger.warning("old version: "+str(oldversion)+" not found in:"+str(oldreportdir))
        return

    logger.info("checking: "+newreportdir)
    logger.info("checking: "+oldreportdir)
    newfiles = {}
    oldfiles = {}
    filelist = []
    #Parse xml files
    for newfile in os.listdir(newreportdir):
        if newfile.find("Schema"):
            continue
        ind = newfile.find("__")
        fileconfig = newfile[ind+2:-4].strip()
        #if fileconfig.find("Diff_"):
        #    pass
        if fileconfig == newconfig:
            reportfile = os.path.join(newreportdir, newfile)
            try:
                newdoc = xml.dom.minidom.parse(reportfile)
            except IOError:
                newdoc = None
                logger.warning("attempted to parse directory in "+newreportdir)
        elif str(newfile).endswith( ".xml" ):
            reportfile = os.path.join(newreportdir, newfile)
            try:
                newdoc = xml.dom.minidom.parse(reportfile)
            except:
                newdoc = None
                pass

        if newdoc:
            newfiles[str(newfile)] = newdoc
            filelist += [newfile]
    #go through old fir

    # Correct for looking for -pre-pre folders
    import os.path
    if not os.path.exists( oldreportdir ):
        if os.path.exists( oldreportdir[:-3] ):
            oldreportdir = oldreportdir[:-3]
        elif os.path.exists( oldreportdir[:-6] ):
            oldreportdir = oldreportdir[:-6]

    for oldfile in os.listdir(oldreportdir):
        if oldfile.find("Schema"):
            continue
        reportfile = os.path.join(oldreportdir,oldfile)
        ind = oldfile.find("__")
        fileconfig = oldfile[ind+2:-4].strip()
        if fileconfig == oldconfig:
            try:
                olddoc = xml.dom.minidom.parse(reportfile)
            except IOError:
                olddoc = None
                logger.warning("attempted to parse directory in "+oldreportdir)     
        elif str(oldfile).endswith( ".xml" ):
            try:
                olddoc = xml.dom.minidom.parse(reportfile)
            except IOError:
                olddoc = None

        if olddoc is not None:
            oldfiles[str(oldfile)] = olddoc

    logger.info("newfiles: " + str(newfiles))
    logger.info("oldfiles: " + str(oldfiles))

    #go through dictionaries and compare files and then compare tests
    comparetestfiles(newfiles, oldfiles)

    logger.info('Tests finished.')
    

def main(cmd_args):
    """
    cmd_args is a list contining the directories holiding the xml files to compare
    """
    start(cmd_args) 




################################################################################
# Ganga Project. http://cern.ch/ganga
#
################################################################################

""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""

external_packages = { }
from GangaCore.Utility.Setup import PackageSetup

setup = PackageSetup(external_packages)

def standardSetup(setup=setup):
    import GangaCore.Utility.Setup
#    GangaCore.Utility.Setup.setPlatform('slc3_gcc323')
    
    for p in setup.packages:
        setup.prependPath(p,'PYTHONPATH')
        setup.prependPath(p,'PATH')
        setup.prependPath(p,'LD_LIBRARY_PATH')
 
